"""
Module for petitCODE atmospheric model spectra.
"""

import os
import tarfile
import urllib.request
import warnings
import zipfile

from typing import Optional, Tuple

import h5py
import numpy as np
import spectres

from typeguard import typechecked

from species.core import constants
from species.util import data_util, read_util


@typechecked
def add_petitcode_cool_clear(input_path: str,
                             database: h5py._hl.files.File,
                             wavel_range: Optional[Tuple[float, float]] = None,
                             teff_range: Optional[Tuple[float, float]] = None,
                             spec_res: Optional[float] = 1000.) -> None:
    """
    Function for adding the petitCODE cool clear atmospheric models to the database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.
    wavel_range : tuple(float, float), None
        Wavelength range (um). The original wavelength points are used if set to None.
    teff_range : tuple(float, float), None
        Effective temperature range (K). All temperatures are selected if set to None.
    spec_res : float, None
        Spectral resolution. Not used if ``wavel_range`` is set to None.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    data_folder = os.path.join(input_path, 'linder_molliere_grid/clear/specs/')

    url = 'http://mpia.de/~molliere/online_data/linder_molliere_grid.zip'

    data_file = os.path.join(input_path, 'linder_molliere_grid.zip')

    if not os.path.isfile(data_file):
        print('Downloading petitCODE cool model spectra (3.7 GB)...', end='', flush=True)
        urllib.request.urlretrieve(url, data_file)
        print(' [DONE]')

    print('Unpacking petitCODE cool model spectra (3.7 GB)...', end='', flush=True)

    with zipfile.ZipFile(data_file, 'r') as zip_ref:
        zip_ref.extractall(input_path)

    print(' [DONE]')

    teff = []
    logg = []
    feh = []
    flux = []

    if wavel_range is not None:
        wavelength = read_util.create_wavelengths(wavel_range, spec_res)
    else:
        wavelength = None

    for _, _, files in os.walk(data_folder):
        for filename in files:
            file_split = filename.split('_')

            teff_val = float(file_split[2])
            logg_val = float(file_split[4])
            feh_val = float(file_split[6])

            if teff_range is not None:
                if teff_val < teff_range[0] or teff_val > teff_range[1]:
                    continue

            print_message = f'Adding petitCODE cool clear model spectra... {filename}'
            print(f'\r{print_message:<87}', end='')

            data = np.loadtxt(os.path.join(data_folder, filename))

            teff.append(teff_val)
            logg.append(logg_val)
            feh.append(feh_val)

            if wavel_range is None:
                if wavelength is None:
                    # (cm) -> (um)
                    wavelength = data[:, 0]*1e4

                if np.all(np.diff(wavelength) < 0):
                    raise ValueError('The wavelengths are not all sorted by increasing value.')

                # (erg s-1 cm-2 Hz-1) -> (W m-2 um-1)
                flux.append(data[:, 1]*1e-9*constants.LIGHT/(wavelength*1e-6)**2)

            else:
                # (cm) -> (um)
                data_wavel = data[:, 0]*1e4

                # (erg s-1 cm-2 Hz-1) -> (W m-2 um-1)
                data_flux = data[:, 1]*1e-9*constants.LIGHT/(data_wavel*1e-6)**2

                try:
                    flux.append(spectres.spectres(wavelength, data_wavel, data_flux))
                except ValueError:
                    flux.append(np.zeros(wavelength.shape[0]))

                    warnings.warn('The wavelength range should fall within the range of the '
                                  'original wavelength sampling. Storing zeros instead.')

    print_message = 'Adding petitCODE cool clear model spectra... [DONE]'
    print(f'\r{print_message:<87}')

    data_sorted = data_util.sort_data(np.asarray(teff),
                                      np.asarray(logg),
                                      np.asarray(feh),
                                      None,
                                      None,
                                      wavelength,
                                      np.asarray(flux))

    data_util.write_data('petitcode-cool-clear',
                         ['teff', 'logg', 'feh'],
                         database,
                         data_sorted)


@typechecked
def add_petitcode_cool_cloudy(input_path: str,
                              database: h5py._hl.files.File,
                              wavel_range: Optional[Tuple[float, float]] = None,
                              teff_range: Optional[Tuple[float, float]] = None,
                              spec_res: Optional[float] = 1000.) -> None:
    """
    Function for adding the petitCODE cool cloudy atmospheric models to the database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.
    wavel_range : tuple(float, float), None
        Wavelength range (um). The original wavelength points are used if set to None.
    teff_range : tuple(float, float), None
        Effective temperature range (K). All temperatures are selected if set to None.
    spec_res : float, None
        Spectral resolution. Not used if ``wavel_range`` is set to None.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    data_folder = os.path.join(input_path, 'linder_molliere_grid/cloudy/specs/')

    url = 'http://mpia.de/~molliere/online_data/linder_molliere_grid.zip'

    data_file = os.path.join(input_path, 'linder_molliere_grid.zip')

    if not os.path.isfile(data_file):
        print('Downloading petitCODE cool model spectra (3.7 GB)...', end='', flush=True)
        urllib.request.urlretrieve(url, data_file)
        print(' [DONE]')

    print('Unpacking petitCODE cool model spectra (3.7 GB)...', end='', flush=True)

    with zipfile.ZipFile(data_file, 'r') as zip_ref:
        zip_ref.extractall(input_path)

    print(' [DONE]')

    teff = []
    logg = []
    feh = []
    fsed = []
    flux = []

    if wavel_range is not None:
        wavelength = read_util.create_wavelengths(wavel_range, spec_res)
    else:
        wavelength = None

    for _, _, files in os.walk(data_folder):
        for filename in files:
            file_split = filename.split('_')

            teff_val = float(file_split[2])
            logg_val = float(file_split[4])
            feh_val = float(file_split[6])
            fsed_val = float(file_split[8])

            if teff_range is not None:
                if teff_val < teff_range[0] or teff_val > teff_range[1]:
                    continue

            print_message = f'Adding petitCODE cool cloudy model spectra... {filename}'
            print(f'\r{print_message:<106}', end='')

            data = np.loadtxt(os.path.join(data_folder, filename))

            teff.append(teff_val)
            logg.append(logg_val)
            feh.append(feh_val)
            fsed.append(fsed_val)

            if wavel_range is None:
                if wavelength is None:
                    # (cm) -> (um)
                    wavelength = data[:, 0]*1e4

                if np.all(np.diff(wavelength) < 0):
                    raise ValueError('The wavelengths are not all sorted by increasing value.')

                # (erg s-1 cm-2 Hz-1) -> (W m-2 um-1)
                flux.append(data[:, 1]*1e-9*constants.LIGHT/(wavelength*1e-6)**2)

            else:
                # (cm) -> (um)
                data_wavel = data[:, 0]*1e4

                # (erg s-1 cm-2 Hz-1) -> (W m-2 um-1)
                data_flux = data[:, 1]*1e-9*constants.LIGHT/(data_wavel*1e-6)**2

                try:
                    flux.append(spectres.spectres(wavelength, data_wavel, data_flux))
                except ValueError:
                    flux.append(np.zeros(wavelength.shape[0]))

                    warnings.warn('The wavelength range should fall within the range of the '
                                  'original wavelength sampling. Storing zeros instead.')

    print_message = 'Adding petitCODE cool cloudy model spectra... [DONE]'
    print(f'\r{print_message:<106}')

    data_sorted = data_util.sort_data(np.asarray(teff),
                                      np.asarray(logg),
                                      np.asarray(feh),
                                      None,
                                      np.asarray(fsed),
                                      wavelength,
                                      np.asarray(flux))

    data_util.write_data('petitcode-cool-cloudy',
                         ['teff', 'logg', 'feh', 'fsed'],
                         database,
                         data_sorted)


@typechecked
def add_petitcode_hot_clear(input_path: str,
                            database: h5py._hl.files.File,
                            wavel_range: Optional[Tuple[float, float]] = None,
                            teff_range: Optional[Tuple[float, float]] = None,
                            spec_res: Optional[float] = 1000.) -> None:
    """
    Function for adding the petitCODE hot clear atmospheric models to the database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.
    wavel_range : tuple(float, float), None
        Wavelength range (um). The original wavelength points are used if set to None.
    teff_range : tuple(float, float), None
        Effective temperature range (K). All temperatures are selected if set to None.
    spec_res : float, None
        Spectral resolution. Not used if ``wavel_range`` is set to None.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    data_folder = os.path.join(input_path, 'petitcode-hot-clear/')

    url = 'https://people.phys.ethz.ch/~ipa/tstolker/petitcode-hot-clear.tgz'

    data_file = os.path.join(input_path, 'petitcode-hot-clear.tgz')

    if not os.path.isfile(data_file):
        print('Downloading petitCODE hot clear model spectra (93 MB)...', end='', flush=True)
        urllib.request.urlretrieve(url, data_file)
        print(' [DONE]')

    print('Unpacking petitCODE hot clear model spectra (93 MB)...', end='', flush=True)
    tar = tarfile.open(data_file)
    tar.extractall(data_folder)
    tar.close()
    print(' [DONE]')

    teff = []
    logg = []
    feh = []
    co_ratio = []
    flux = []

    if wavel_range is not None:
        wavelength = read_util.create_wavelengths(wavel_range, spec_res)
    else:
        wavelength = None

    for _, _, files in os.walk(data_folder):
        for filename in files:
            file_split = filename.split('_')

            teff_val = float(file_split[2])
            logg_val = float(file_split[4])
            feh_val = float(file_split[6])
            co_ratio_val = float(file_split[8])

            if teff_range is not None:
                if teff_val < teff_range[0] or teff_val > teff_range[1]:
                    continue

            print_message = f'Adding petitCODE hot clear model spectra... {filename}'
            print(f'\r{print_message:<99}', end='')

            data = np.loadtxt(os.path.join(data_folder, filename))

            teff.append(teff_val)
            logg.append(logg_val)
            feh.append(feh_val)
            co_ratio.append(co_ratio_val)

            if wavel_range is None:
                if wavelength is None:
                    # (cm) -> (um)
                    wavelength = data[:, 0]*1e4

                if np.all(np.diff(wavelength) < 0):
                    raise ValueError('The wavelengths are not all sorted by increasing value.')

                # (erg s-1 cm-2 Hz-1) -> (W m-2 um-1)
                flux.append(data[:, 1]*1e-9*constants.LIGHT/(wavelength*1e-6)**2)

            else:
                # (cm) -> (um)
                data_wavel = data[:, 0]*1e4

                # (erg s-1 cm-2 Hz-1) -> (W m-2 um-1)
                data_flux = data[:, 1]*1e-9*constants.LIGHT/(data_wavel*1e-6)**2

                try:
                    flux.append(spectres.spectres(wavelength, data_wavel, data_flux))
                except ValueError:
                    flux.append(np.zeros(wavelength.shape[0]))

                    warnings.warn('The wavelength range should fall within the range of the '
                                  'original wavelength sampling. Storing zeros instead.')

    print_message = 'Adding petitCODE hot clear model spectra... [DONE]'
    print(f'\r{print_message:<99}')

    data_sorted = data_util.sort_data(np.asarray(teff),
                                      np.asarray(logg),
                                      np.asarray(feh),
                                      np.asarray(co_ratio),
                                      None,
                                      wavelength,
                                      np.asarray(flux))

    data_util.write_data('petitcode-hot-clear',
                         ['teff', 'logg', 'feh', 'co'],
                         database,
                         data_sorted)


@typechecked
def add_petitcode_hot_cloudy(input_path: str,
                             database: h5py._hl.files.File,
                             wavel_range: Optional[Tuple[float, float]] = None,
                             teff_range: Optional[Tuple[float, float]] = None,
                             spec_res: Optional[float] = 1000.) -> None:
    """
    Function for adding the petitCODE hot cloudy atmospheric models to the database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.
    wavel_range : tuple(float, float), None
        Wavelength range (um). The original wavelength points are used if set to None.
    teff_range : tuple(float, float), None
        Effective temperature range (K). All temperatures are selected if set to None.
    spec_res : float, None
        Spectral resolution. Not used if ``wavel_range`` is set to None.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    data_folder = os.path.join(input_path, 'petitcode-hot-cloudy/')

    url = 'https://people.phys.ethz.ch/~ipa/tstolker/petitcode-hot-cloudy.tgz'

    data_file = os.path.join(input_path, 'petitcode-hot-cloudy.tgz')

    if not os.path.isfile(data_file):
        print('Downloading petitCODE hot cloudy model spectra (276 MB)...', end='', flush=True)
        urllib.request.urlretrieve(url, data_file)
        print(' [DONE]')

    print('Unpacking petitCODE hot cloudy model spectra (276 MB)...', end='', flush=True)
    tar = tarfile.open(data_file)
    tar.extractall(data_folder)
    tar.close()
    print(' [DONE]')

    teff = []
    logg = []
    feh = []
    co_ratio = []
    fsed = []
    flux = []

    if wavel_range is not None:
        wavelength = read_util.create_wavelengths(wavel_range, spec_res)
    else:
        wavelength = None

    for _, _, files in os.walk(data_folder):
        for filename in files:
            file_split = filename.split('_')

            teff_val = float(file_split[2])
            logg_val = float(file_split[4])
            feh_val = float(file_split[6])
            co_ratio_val = float(file_split[8])
            fsed_val = float(file_split[10])

            if teff_range is not None:
                if teff_val < teff_range[0] or teff_val > teff_range[1]:
                    continue

            print_message = f'Adding petitCODE hot cloudy model spectra... {filename}'
            print(f'\r{print_message:<111}', end='')

            data = np.loadtxt(os.path.join(data_folder, filename))

            teff.append(teff_val)
            logg.append(logg_val)
            feh.append(feh_val)
            co_ratio.append(co_ratio_val)
            fsed.append(fsed_val)

            if wavel_range is None:
                if wavelength is None:
                    # (cm) -> (um)
                    wavelength = data[:, 0]*1e4

                if np.all(np.diff(wavelength) < 0):
                    raise ValueError('The wavelengths are not all sorted by increasing value.')

                # (erg s-1 cm-2 Hz-1) -> (W m-2 um-1)
                flux.append(data[:, 1]*1e-9*constants.LIGHT/(wavelength*1e-6)**2)

            else:
                # (cm) -> (um)
                data_wavel = data[:, 0]*1e4

                # (erg s-1 cm-2 Hz-1) -> (W m-2 um-1)
                data_flux = data[:, 1]*1e-9*constants.LIGHT/(data_wavel*1e-6)**2

                try:
                    flux.append(spectres.spectres(wavelength, data_wavel, data_flux))
                except ValueError:
                    flux.append(np.zeros(wavelength.shape[0]))

                    warnings.warn('The wavelength range should fall within the range of the '
                                  'original wavelength sampling. Storing zeros instead.')

    print_message = 'Adding petitCODE hot cloudy model spectra... [DONE]'
    print(f'\r{print_message:<111}')

    data_sorted = data_util.sort_data(np.asarray(teff),
                                      np.asarray(logg),
                                      np.asarray(feh),
                                      np.asarray(co_ratio),
                                      np.asarray(fsed),
                                      wavelength,
                                      np.asarray(flux))

    data_util.write_data('petitcode-hot-cloudy',
                         ['teff', 'logg', 'feh', 'co', 'fsed'],
                         database,
                         data_sorted)



"""
Module with functions for creating color-magnitude and color-color plot.
"""

import math
import os

from typing import List, Optional, Tuple, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from matplotlib.colorbar import Colorbar
from matplotlib.ticker import MultipleLocator
from scipy.interpolate import interp1d
from typeguard import typechecked

from species.core import box
from species.read import read_object
from species.util import dust_util, plot_util


@typechecked
def plot_color_magnitude(boxes: list,
                         objects: Optional[Union[List[Tuple[str, str, str, str]],
                                                 List[Tuple[str, str, str, str, Optional[dict],
                                                            Optional[dict]]]]] = None,
                         mass_labels: Optional[Union[List[float], List[Tuple[float, str]]]] = None,
                         teff_labels: Optional[Union[List[float], List[Tuple[float, str]]]] = None,
                         companion_labels: bool = False,
                         reddening: Optional[List[Tuple[Tuple[str, str], Tuple[str, float], str,
                                                        float, Tuple[float, float]]]] = None,
                         field_range: Optional[Tuple[str, str]] = None,
                         label_x: str = 'Color',
                         label_y: str = 'Absolute magnitude',
                         xlim: Optional[Tuple[float, float]] = None,
                         ylim: Optional[Tuple[float, float]] = None,
                         offset: Optional[Tuple[float, float]] = None,
                         legend: Optional[Union[str, dict, Tuple[float, float]]] = 'upper left',
                         output: str = 'color-magnitude.pdf') -> None:
    """
    Function for creating a color-magnitude diagram.

    Parameters
    ----------
    boxes : list(species.core.box.ColorMagBox, species.core.box.IsochroneBox, )
        Boxes with the color-magnitude and isochrone data from photometric libraries, spectral
        libraries, and/or atmospheric models. The synthetic data have to be created with
        :func:`~species.read.read_isochrone.ReadIsochrone.get_color_magnitude`. These boxes
        contain synthetic colors and magnitudes for a given age and a range of masses.
    objects : list(tuple(str, str, str, str), ),
              list(tuple(str, str, str, str, dict, dict), ), None
        Tuple with individual objects. The objects require a tuple with their database tag, the two
        filter names for the color, and the filter name for the absolute magnitude. Optionally, a
        dictionary with keyword arguments can be provided for the object's marker and label,
        respectively. For example, ``{'marker': 'o', 'ms': 10}`` for the marker and
        ``{'ha': 'left', 'va': 'bottom', 'xytext': (5, 5)})`` for the label. Not used if set to
        None.
    mass_labels : list(float, ), list(tuple(float, str), ), None
        Plot labels with masses next to the isochrone data of `models`. The list with masses has
        to be provided in Jupiter mass. Alternatively, a list of tuples can be provided with
        the planet mass and position of the label ('left' or 'right), for example
        ``[(10., 'left'), (20., 'right')]``. No labels are shown if set to None.
    teff_labels : list(float, ), list(tuple(float, str), ), None
        Plot labels with temperatures (K) next to the synthetic Planck photometry. Alternatively,
        a list of tuples can be provided with the planet mass and position of the label ('left' or
        'right), for example ``[(1000., 'left'), (1200., 'right')]``. No labels are shown if set
        to None.
    companion_labels : bool
        Plot labels with the names of the directly imaged companions.
    reddening : list(tuple(tuple(str, str), tuple(str, float), str, float, tuple(float, float)), None
        Include reddening arrows by providing a list with tuples. Each tuple contains the filter
        names for the color, the filter name and value of the magnitude, the particle radius (um),
        and the start position (color, mag) of the arrow in the plot, so ``((filter_color_1,
        filter_color_2), (filter_mag, mag_value), composition, radius, (x_pos, y_pos))``. The
        composition can be either ``'Fe'`` or ``'MgSiO3'`` (both with crystalline structure). Both
        ``xlim`` and ``ylim`` need to be set for the correct rotation of the reddening label. The
        parameter is not used if set to ``None``.
    field_range : tuple(str, str), None
        Range of the discrete colorbar for the field dwarfs. The tuple should contain the lower
        and upper value ('early M', 'late M', 'early L', 'late L', 'early T', 'late T', 'early Y).
        The full range is used if set to None.
    label_x : str
        Label for the x-axis.
    label_y : str
        Label for the y-axis.
    xlim : tuple(float, float), None
        Limits for the x-axis. Not used if set to None.
    ylim : tuple(float, float), None
        Limits for the y-axis. Not used if set to None.
    offset : tuple(float, float), None
        Offset of the x- and y-axis label.
    legend : str, tuple(float, float), dict, None
        Legend position or keyword arguments. No legend is shown if set to ``None``.
    output : str
        Output filename.

    Returns
    -------
    NoneType
        None

    """

    mpl.rcParams['font.serif'] = ['Bitstream Vera Serif']
    mpl.rcParams['font.family'] = 'serif'

    plt.rc('axes', edgecolor='black', linewidth=2.2)

    model_color = ('#234398', '#f6a432', 'black')
    model_linestyle = ('-', '--', ':', '-.')

    isochrones = []
    planck = []
    models = []
    empirical = []

    for item in boxes:
        if isinstance(item, box.IsochroneBox):
            isochrones.append(item)

        elif isinstance(item, box.ColorMagBox):
            if item.object_type == 'model':
                models.append(item)

            elif item.library == 'planck':
                planck.append(item)

            else:
                empirical.append(item)

        else:
            raise ValueError(f'Found a {type(item)} while only ColorMagBox and IsochroneBox '
                             f'objects can be provided to \'boxes\'.')

    if empirical:
        plt.figure(1, figsize=(4., 4.8))
        gridsp = mpl.gridspec.GridSpec(3, 1, height_ratios=[0.2, 0.1, 4.5])
        gridsp.update(wspace=0., hspace=0., left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[2, 0])
        ax2 = plt.subplot(gridsp[0, 0])

    else:
        plt.figure(1, figsize=(4., 4.5))
        gridsp = mpl.gridspec.GridSpec(1, 1)
        gridsp.update(wspace=0., hspace=0., left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[0, 0])

    ax1.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                    direction='in', width=1, length=5, labelsize=12, top=True,
                    bottom=True, left=True, right=True)

    ax1.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                    direction='in', width=1, length=3, labelsize=12, top=True,
                    bottom=True, left=True, right=True)

    ax1.xaxis.set_major_locator(MultipleLocator(1.))
    ax1.yaxis.set_major_locator(MultipleLocator(1.))

    ax1.xaxis.set_minor_locator(MultipleLocator(0.2))
    ax1.yaxis.set_minor_locator(MultipleLocator(0.2))

    ax1.set_xlabel(label_x, fontsize=14)
    ax1.set_ylabel(label_y, fontsize=14)

    ax1.invert_yaxis()

    if offset is not None:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])
        ax1.get_yaxis().set_label_coords(offset[1], 0.5)

    else:
        ax1.get_xaxis().set_label_coords(0.5, -0.08)
        ax1.get_yaxis().set_label_coords(-0.12, 0.5)

    if xlim is not None:
        ax1.set_xlim(xlim[0], xlim[1])

    if ylim is not None:
        ax1.set_ylim(ylim[0], ylim[1])

    if models is not None:
        count = 0

        model_dict = {}

        for j, item in enumerate(models):
            if item.library not in model_dict:
                model_dict[item.library] = [count, 0]
                count += 1

            else:
                model_dict[item.library] = [model_dict[item.library][0],
                                            model_dict[item.library][1]+1]

            model_count = model_dict[item.library]

            if model_count[0] == 3:
                raise ValueError('Only three different types of model atmospheres can be added.')

            if model_count[1] == 0:
                label = plot_util.model_name(item.library)

                if item.library == 'zhu2015':
                    ax1.plot(item.color, item.magnitude, marker='x', ms=5, linestyle=model_linestyle[model_count[1]],
                             linewidth=0.6, color='gray', label=label, zorder=0)

                    xlim = ax1.get_xlim()
                    ylim = ax1.get_ylim()

                    for i, teff_item in enumerate(item.sptype):
                        teff_label = rf'{teff_item:.0e} $M_\mathregular{{Jup}}^{2}$ yr$^{{-1}}$'

                        if item.magnitude[i] > ylim[1]:
                            ax1.annotate(teff_label, (item.color[i], item.magnitude[i]),
                                         color='gray', fontsize=8, ha='left', va='center',
                                         xytext=(item.color[i]+0.1, item.magnitude[i]+0.05), zorder=3)

                else:
                    ax1.plot(item.color, item.magnitude, linestyle=model_linestyle[model_count[1]],
                             linewidth=1., color=model_color[model_count[0]], label=label, zorder=0)

                    if mass_labels is not None:
                        interp_magnitude = interp1d(item.sptype, item.magnitude)
                        interp_color = interp1d(item.sptype, item.color)

                        for i, mass_item in enumerate(mass_labels):
                            if isinstance(mass_item, tuple):
                                mass_val = mass_item[0]
                                mass_pos = mass_item[1]

                            else:
                                mass_val = mass_item
                                mass_pos = 'right'

                            if j == 0 or (j > 0 and mass_val < 20.):
                                pos_color = interp_color(mass_val)
                                pos_mag = interp_magnitude(mass_val)

                                if mass_pos == 'left':
                                    mass_ha = 'right'
                                    mass_xytext = (pos_color-0.05, pos_mag)

                                else:
                                    mass_ha = 'left'
                                    mass_xytext = (pos_color+0.05, pos_mag)

                                mass_label = str(int(mass_val))+r' M$_\mathregular{J}$'

                                xlim = ax1.get_xlim()
                                ylim = ax1.get_ylim()

                                if xlim[0]+0.2 < pos_color < xlim[1]-0.2 and \
                                        ylim[1]+0.2 < pos_mag < ylim[0]-0.2:

                                    ax1.scatter(pos_color, pos_mag, c=model_color[model_count[0]], s=15,
                                                edgecolor='none', zorder=0)

                                    ax1.annotate(mass_label, (pos_color, pos_mag),
                                                 color=model_color[model_count[0]], fontsize=9,
                                                 xytext=mass_xytext, zorder=3, ha=mass_ha, va='center')

            else:
                ax1.plot(item.color, item.magnitude, linestyle=model_linestyle[model_count[1]],
                         linewidth=0.6, color=model_color[model_count[0]], zorder=0)

    if planck is not None:
        planck_count = 0

        for j, item in enumerate(planck):
            if planck_count == 0:
                label = plot_util.model_name(item.library)
            else:
                label = None

            ax1.plot(item.color, item.magnitude, linestyle='--', linewidth=0.8,
                     color='black', label=label, zorder=0)

            if teff_labels is not None and planck_count == 0:
                interp_magnitude = interp1d(item.sptype, item.magnitude)
                interp_color = interp1d(item.sptype, item.color)

                for i, teff_item in enumerate(teff_labels):
                    if isinstance(teff_item, tuple):
                        teff_val = teff_item[0]
                        teff_pos = teff_item[1]

                    else:
                        teff_val = teff_item
                        teff_pos = 'right'

                    if j == 0 or (j > 0 and teff_val < 20.):
                        pos_color = interp_color(teff_val)
                        pos_mag = interp_magnitude(teff_val)

                        if teff_pos == 'left':
                            teff_ha = 'right'
                            teff_xytext = (pos_color-0.05, pos_mag)

                        else:
                            teff_ha = 'left'
                            teff_xytext = (pos_color+0.05, pos_mag)

                        teff_label = f'{int(teff_val)} K'

                        xlim = ax1.get_xlim()
                        ylim = ax1.get_ylim()

                        if xlim[0]+0.2 < pos_color < xlim[1]-0.2 and \
                                ylim[1]+0.2 < pos_mag < ylim[0]-0.2:

                            ax1.scatter(pos_color, pos_mag, c='black', s=15, ec='none', zorder=0)

                            if planck_count == 0:
                                ax1.annotate(teff_label, (pos_color, pos_mag), color='black',
                                             fontsize=9, xytext=teff_xytext, zorder=3, ha=teff_ha,
                                             va='center')

            planck_count += 1

    if empirical:
        cmap = plt.cm.viridis

        bounds, ticks, ticklabels = plot_util.field_bounds_ticks(field_range)
        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)

        for item in empirical:
            sptype = item.sptype
            color = item.color
            magnitude = item.magnitude
            names = item.names

            if isinstance(sptype, list):
                sptype = np.array(sptype)

            if item.object_type in ['field', None]:
                indices = np.where(sptype != 'None')[0]

                sptype = sptype[indices]
                color = color[indices]
                magnitude = magnitude[indices]

                spt_disc = plot_util.sptype_substellar(sptype, color.shape)

                _, unique = np.unique(color, return_index=True)

                sptype = sptype[unique]
                color = color[unique]
                magnitude = magnitude[unique]
                spt_disc = spt_disc[unique]

                scat = ax1.scatter(color, magnitude, c=spt_disc, cmap=cmap, norm=norm, s=50,
                                   alpha=0.7, edgecolor='none', zorder=2)

                cb = Colorbar(ax=ax2, mappable=scat, orientation='horizontal',
                              ticklocation='top', format='%.2f')

                cb.ax.tick_params(width=1, length=5, labelsize=10, direction='in', color='black')

                cb.set_ticks(ticks)
                cb.set_ticklabels(ticklabels)

            elif item.object_type == 'young':
                if objects is not None:
                    object_names = []

                    for obj_item in objects:
                        object_names.append(obj_item[0])

                    indices = plot_util.remove_color_duplicates(object_names, names)

                    color = color[indices]
                    magnitude = magnitude[indices]

                ax1.plot(color, magnitude, marker='s', ms=4, linestyle='none', alpha=0.7,
                         color='gray', markeredgecolor='black', label='Young/low-gravity', zorder=2)

    if isochrones:
        for item in isochrones:
            ax1.plot(item.color, item.magnitude, linestyle='-', linewidth=1., color='black')

    if reddening is not None:
        for item in reddening:
            ext_1, ext_2 = dust_util.calc_reddening(item[0],
                                                    item[1],
                                                    composition=item[2],
                                                    structure='crystalline',
                                                    radius_g=item[3])

            delta_x = ext_1 - ext_2
            delta_y = item[1][1]

            x_pos = item[4][0] + delta_x
            y_pos = item[4][1] + delta_y

            ax1.annotate('', (x_pos, y_pos), xytext=(item[4][0], item[4][1]),
                         fontsize=8, arrowprops={'arrowstyle': '->'}, color='black', zorder=3.)

            x_pos_text = item[4][0] + delta_x/2.
            y_pos_text = item[4][1] + delta_y/2.

            vector_len = math.sqrt(delta_x**2+delta_y**2)

            if item[2] == 'MgSiO3':
                dust_species = r'MgSiO$_{3}$'
            elif item[2] == 'Fe':
                dust_species = 'Fe'

            if (item[3]).is_integer():
                red_label = f'{dust_species} ({item[3]:.0f} µm)'
            else:
                red_label = f'{dust_species} ({item[3]:.1f} µm)'

            text = ax1.annotate(red_label, (x_pos_text, y_pos_text),
                                xytext=(7.*delta_y/vector_len, 7.*delta_x/vector_len),
                                textcoords='offset points', fontsize=8., color='black',
                                ha='center', va='center')

            ax1.plot([item[4][0], x_pos], [item[4][1], y_pos], '-', color='white')

            sp1 = ax1.transData.transform_point((item[4][0], item[4][1]))
            sp2 = ax1.transData.transform_point((x_pos, y_pos))

            angle = np.degrees(np.arctan2(sp2[1]-sp1[1], sp2[0]-sp1[0]))
            text.set_rotation(angle)

    if objects is not None:
        for i, item in enumerate(objects):
            objdata = read_object.ReadObject(item[0])

            objcolor1 = objdata.get_photometry(item[1])
            objcolor2 = objdata.get_photometry(item[2])

            if objcolor1.ndim == 2:
                print(f'Found {objcolor1.shape[1]} values for filter {item[1]} of {item[0]}')
                print(f'so using the first value:  {objcolor1[0, 0]} +/- {objcolor1[1, 0]} mag')
                objcolor1 = objcolor1[:, 0]

            if objcolor2.ndim == 2:
                print(f'Found {objcolor2.shape[1]} values for filter {item[2]} of {item[0]}')
                print(f'so using the first value:  {objcolor2[0, 0]} +/- {objcolor2[1, 0]} mag')
                objcolor2 = objcolor2[:, 0]

            abs_mag, abs_err = objdata.get_absmag(item[3])

            if isinstance(abs_mag, np.ndarray):
                abs_mag = abs_mag[0]
                abs_err = abs_err[0]

            colorerr = math.sqrt(objcolor1[1]**2+objcolor2[1]**2)
            x_color = objcolor1[0]-objcolor2[0]

            if len(item) > 4 and item[4] is not None:
                kwargs = item[4]

            else:
                kwargs = {'marker': '>',
                          'ms': 6.,
                          'color': 'black',
                          'mfc': 'white',
                          'mec': 'black',
                          'label': 'Direct imaging'}

            ax1.errorbar(x_color, abs_mag, yerr=abs_err, xerr=colorerr, zorder=3, **kwargs)

            if companion_labels:
                if len(item) > 4:
                    kwargs = item[5]

                else:
                    kwargs = {'ha': 'left',
                              'va': 'bottom',
                              'fontsize': 8.5,
                              'xytext': (5., 5.),
                              'color': 'black'}

                ax1.annotate(objdata.object_name, (x_color, abs_mag), zorder=3,
                             textcoords='offset points', **kwargs)

    print(f'Plotting color-magnitude diagram: {output}...', end='', flush=True)

    if legend is not None:
        handles, labels = ax1.get_legend_handles_labels()

        # prevent duplicates
        by_label = dict(zip(labels, handles))

        if handles:
            ax1.legend(by_label.values(), by_label.keys(), loc=legend, fontsize=8.5,
                       frameon=False, numpoints=1)

    plt.savefig(os.getcwd()+'/'+output, bbox_inches='tight')
    plt.clf()
    plt.close()

    print(' [DONE]')


@typechecked
def plot_color_color(boxes: list,
                     objects: Optional[Union[List[Tuple[str, Tuple[str, str], Tuple[str, str]]],
                                             List[Tuple[str, Tuple[str, str], Tuple[str, str], Optional[dict],
                                                        Optional[dict]]]]] = None,
                     mass_labels: Optional[Union[List[float], List[Tuple[float, str]]]] = None,
                     teff_labels: Optional[Union[List[float], List[Tuple[float, str]]]] = None,
                     companion_labels: bool = False,
                     reddening: Optional[List[Tuple[Tuple[str, str], Tuple[str, str],
                                                    Tuple[str, float], str, float,
                                                    Tuple[float, float]]]] = None,
                     field_range: Optional[Tuple[str, str]] = None,
                     label_x: str = 'Color',
                     label_y: str = 'Color',
                     xlim: Optional[Tuple[float, float]] = None,
                     ylim: Optional[Tuple[float, float]] = None,
                     offset: Optional[Tuple[float, float]] = None,
                     legend: Optional[Union[str, dict, Tuple[float, float]]] = 'upper left',
                     figsize: Optional[Tuple[float, float]] = (4., 4.3),
                     output: str = 'color-color.pdf') -> None:
    """
    Function for creating a color-color diagram.

    Parameters
    ----------
    boxes : list(species.core.box.ColorColorBox, species.core.box.IsochroneBox, )
        Boxes with the color-color and isochrone data from photometric libraries, spectral
        libraries, and/or atmospheric models. The synthetic data have to be created with
        :func:`~species.read.read_isochrone.ReadIsochrone.get_color_color`. These boxes
        contain synthetic colors for a given age and a range of masses.
    objects : tuple(tuple(str, tuple(str, str), tuple(str, str)), ),
              tuple(tuple(str, tuple(str, str), tuple(str, str), dict, dict), ), None
        Tuple with individual objects. The objects require a tuple with their database tag, the two
        filter names for the first color, and the two filter names for the second color.
        Optionally, a dictionary with keyword arguments can be provided for the object's marker and
        label, respectively. For example, ``{'marker': 'o', 'ms': 10}`` for the marker and
        ``{'ha': 'left', 'va': 'bottom', 'xytext': (5, 5)})`` for the label. Not used if set to
        None.
    mass_labels : list(float, ), list(tuple(float, str), ), None
        Plot labels with masses next to the isochrone data of `models`. The list with masses has
        to be provided in Jupiter mass. Alternatively, a list of tuples can be provided with
        the planet mass and position of the label ('left' or 'right), for example
        ``[(10., 'left'), (20., 'right')]``. No labels are shown if set to None.
    teff_labels : list(float, ), list(tuple(float, str), ), None
        Plot labels with temperatures (K) next to the synthetic Planck photometry. Alternatively,
        a list of tuples can be provided with the planet mass and position of the label ('left' or
        'right), for example ``[(1000., 'left'), (1200., 'right')]``. No labels are shown if set
        to None.
    companion_labels : bool
        Plot labels with the names of the directly imaged companions.
    reddening : list(tuple(tuple(str, str), tuple(str, str), tuple(str, float), str, float, tuple(float, float)), None
        Include reddening arrows by providing a list with tuples. Each tuple contains the filter
        names for the color, the filter name for the magnitude, the particle radius (um), and the
        start position (color, mag) of the arrow in the plot, so (filter_color_1, filter_color_2,
        filter_mag, composition, radius, (x_pos, y_pos)). The composition can be either 'Fe' or
        'MgSiO3' (both with crystalline structure). The parameter is not used if set to ``None``.
    field_range : tuple(str, str), None
        Range of the discrete colorbar for the field dwarfs. The tuple should contain the lower
        and upper value ('early M', 'late M', 'early L', 'late L', 'early T', 'late T', 'early Y).
        The full range is used if set to None.
    label_x : str
        Label for the x-axis.
    label_y : str
        Label for the y-axis.
    xlim : tuple(float, float)
        Limits for the x-axis.
    ylim : tuple(float, float)
        Limits for the y-axis.
    offset : tuple(float, float), None
        Offset of the x- and y-axis label.
    legend : str, tuple(float, float), dict, None
        Legend position or keyword arguments. No legend is shown if set to ``None``.
    figsize : tuple(float, float)
        Figure size.
    output : str
        Output filename.

    Returns
    -------
    NoneType
        None
    """

    mpl.rcParams['font.serif'] = ['Bitstream Vera Serif']
    mpl.rcParams['font.family'] = 'serif'

    plt.rc('axes', edgecolor='black', linewidth=2.2)

    model_color = ('#234398', '#f6a432', 'black')
    model_linestyle = ('-', '--', ':', '-.')

    isochrones = []
    planck = []
    models = []
    empirical = []

    for item in boxes:
        if isinstance(item, box.IsochroneBox):
            isochrones.append(item)

        elif isinstance(item, box.ColorColorBox):
            if item.object_type == 'model':
                models.append(item)

            elif item.library == 'planck':
                planck.append(item)

            else:
                empirical.append(item)

        else:
            raise ValueError(f'Found a {type(item)} while only ColorColorBox and IsochroneBox '
                             f'objects can be provided to \'boxes\'.')

    plt.figure(1, figsize=figsize)
    gridsp = mpl.gridspec.GridSpec(3, 1, height_ratios=[0.2, 0.1, 4.])
    gridsp.update(wspace=0., hspace=0., left=0, right=1, bottom=0, top=1)

    ax1 = plt.subplot(gridsp[2, 0])
    ax2 = plt.subplot(gridsp[0, 0])

    ax1.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                    direction='in', width=1, length=5, labelsize=12, top=True,
                    bottom=True, left=True, right=True)

    ax1.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                    direction='in', width=1, length=3, labelsize=12, top=True,
                    bottom=True, left=True, right=True)

    ax1.xaxis.set_major_locator(MultipleLocator(0.5))
    ax1.yaxis.set_major_locator(MultipleLocator(0.5))

    ax1.xaxis.set_minor_locator(MultipleLocator(0.1))
    ax1.yaxis.set_minor_locator(MultipleLocator(0.1))

    ax1.set_xlabel(label_x, fontsize=14)
    ax1.set_ylabel(label_y, fontsize=14)

    ax1.invert_yaxis()

    if offset:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])
        ax1.get_yaxis().set_label_coords(offset[1], 0.5)
    else:
        ax1.get_xaxis().set_label_coords(0.5, -0.08)
        ax1.get_yaxis().set_label_coords(-0.12, 0.5)

    if xlim:
        ax1.set_xlim(xlim[0], xlim[1])

    if ylim:
        ax1.set_ylim(ylim[0], ylim[1])

    if models is not None:
        count = 0

        model_dict = {}

        for j, item in enumerate(models):
            if item.library not in model_dict:
                model_dict[item.library] = [count, 0]
                count += 1

            else:
                model_dict[item.library] = [model_dict[item.library][0],
                                            model_dict[item.library][1]+1]

            model_count = model_dict[item.library]

            if model_count[1] == 0:
                label = plot_util.model_name(item.library)

                if item.library == 'zhu2015':
                    ax1.plot(item.color1, item.color2, marker='x', ms=5, linestyle=model_linestyle[model_count[1]],
                             linewidth=0.6, color='gray', label=label, zorder=0)

                    xlim = ax1.get_xlim()
                    ylim = ax1.get_ylim()

                    for i, teff_item in enumerate(item.sptype):
                        teff_label = rf'{teff_item:.0e} $M_\mathregular{{Jup}}^{2}$ yr$^{{-1}}$'

                        if item.color2[i] < ylim[1]:
                            ax1.annotate(teff_label, (item.color1[i], item.color2[i]),
                                         color='gray', fontsize=8, ha='left', va='center',
                                         xytext=(item.color1[i]+0.1, item.color2[i]-0.05), zorder=3)

                else:
                    ax1.plot(item.color1, item.color2, linestyle=model_linestyle[model_count[1]],
                             linewidth=1., color=model_color[model_count[0]], label=label, zorder=0)

                    if mass_labels is not None:
                        interp_color1 = interp1d(item.sptype, item.color1)
                        interp_color2 = interp1d(item.sptype, item.color2)

                        for i, mass_item in enumerate(mass_labels):
                            if isinstance(mass_item, tuple):
                                mass_val = mass_item[0]
                                mass_pos = mass_item[1]

                            else:
                                mass_val = mass_item
                                mass_pos = 'right'

                            # if j == 0 or (j > 0 and mass_val < 20.):
                            if j == 0:
                                pos_color1 = interp_color1(mass_val)
                                pos_color2 = interp_color2(mass_val)

                                if mass_pos == 'left':
                                    mass_ha = 'right'
                                    mass_xytext = (pos_color1-0.05, pos_color2)

                                else:
                                    mass_ha = 'left'
                                    mass_xytext = (pos_color1+0.05, pos_color2)

                                mass_label = str(int(mass_val))+r' M$_\mathregular{J}$'

                                xlim = ax1.get_xlim()
                                ylim = ax1.get_ylim()

                                if xlim[0]+0.2 < pos_color1 < xlim[1]-0.2 and \
                                        ylim[0]+0.2 < pos_color2 < ylim[1]-0.2:

                                    ax1.scatter(pos_color1, pos_color2, c=model_color[model_count[0]],
                                                s=15, edgecolor='none', zorder=0)

                                    ax1.annotate(mass_label, (pos_color1, pos_color2),
                                                 color=model_color[model_count[0]], fontsize=9,
                                                 xytext=mass_xytext, ha=mass_ha, va='center', zorder=3)

            else:
                ax1.plot(item.color1, item.color2, linestyle=model_linestyle[model_count[1]],
                         linewidth=0.6, color=model_color[model_count[0]], label=label, zorder=0)

    if planck is not None:
        planck_count = 0

        for j, item in enumerate(planck):

            if planck_count == 0:
                label = plot_util.model_name(item.library)

                ax1.plot(item.color1, item.color2, ls='--', linewidth=0.8, color='black',
                         label=label, zorder=0)

                if teff_labels is not None:
                    interp_color1 = interp1d(item.sptype, item.color1)
                    interp_color2 = interp1d(item.sptype, item.color2)

                    for i, teff_item in enumerate(teff_labels):
                        if isinstance(teff_item, tuple):
                            teff_val = teff_item[0]
                            teff_pos = teff_item[1]

                        else:
                            teff_val = teff_item
                            teff_pos = 'right'

                        if j == 0 or (j > 0 and teff_val < 20.):
                            pos_color1 = interp_color1(teff_val)
                            pos_color2 = interp_color2(teff_val)

                            if teff_pos == 'left':
                                teff_ha = 'right'
                                teff_xytext = (pos_color1-0.05, pos_color2)

                            else:
                                teff_ha = 'left'
                                teff_xytext = (pos_color1+0.05, pos_color2)

                            teff_label = f'{int(teff_val)} K'

                            xlim = ax1.get_xlim()
                            ylim = ax1.get_ylim()

                            if xlim[0]+0.2 < pos_color1 < xlim[1]-0.2 and \
                                    ylim[0]+0.2 < pos_color2 < ylim[1]-0.2:

                                ax1.scatter(pos_color1, pos_color2, c='black', s=15,
                                            edgecolor='none', zorder=0)

                                ax1.annotate(teff_label, (pos_color1, pos_color2),
                                             color='black', fontsize=9,
                                             xytext=teff_xytext, zorder=3, ha=teff_ha, va='center')

            else:
                ax1.plot(item.color1, item.color2, ls='--', lw=0.5, color='black', zorder=0)

            planck_count += 1

    if empirical:
        cmap = plt.cm.viridis

        bounds, ticks, ticklabels = plot_util.field_bounds_ticks(field_range)
        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)

        for item in empirical:
            sptype = item.sptype
            names = item.names
            color1 = item.color1
            color2 = item.color2

            if isinstance(sptype, list):
                sptype = np.array(sptype)

            if item.object_type in ['field', None]:
                indices = np.where(sptype != 'None')[0]

                sptype = sptype[indices]
                color1 = color1[indices]
                color2 = color2[indices]

                spt_disc = plot_util.sptype_substellar(sptype, color1.shape)
                _, unique = np.unique(color1, return_index=True)

                sptype = sptype[unique]
                color1 = color1[unique]
                color2 = color2[unique]
                spt_disc = spt_disc[unique]

                scat = ax1.scatter(color1, color2, c=spt_disc, cmap=cmap, norm=norm, s=50,
                                   alpha=0.7, edgecolor='none', zorder=2)

                cb = Colorbar(ax=ax2, mappable=scat, orientation='horizontal',
                              ticklocation='top', format='%.2f')

                cb.ax.tick_params(width=1, length=5, labelsize=10, direction='in', color='black')

                cb.set_ticks(ticks)
                cb.set_ticklabels(ticklabels)

            elif item.object_type == 'young':
                if objects is not None:
                    object_names = []

                    for obj_item in objects:
                        object_names.append(obj_item[0])

                    indices = plot_util.remove_color_duplicates(object_names, names)

                    color1 = color1[indices]
                    color2 = color2[indices]

                ax1.plot(color1, color2, marker='s', ms=4, linestyle='none', alpha=0.7,
                         color='gray', markeredgecolor='black', label='Young/low-gravity', zorder=2)

    if isochrones:
        for item in isochrones:
            ax1.plot(item.colors[0], item.colors[1], linestyle='-', linewidth=1., color='black')

    if reddening is not None:
        for item in reddening:
            ext_1, ext_2 = dust_util.calc_reddening(item[0],
                                                    item[2],
                                                    composition=item[3],
                                                    structure='crystalline',
                                                    radius_g=item[4])

            ext_3, ext_4 = dust_util.calc_reddening(item[1],
                                                    item[2],
                                                    composition=item[3],
                                                    structure='crystalline',
                                                    radius_g=item[4])

            delta_x = ext_1 - ext_2
            delta_y = ext_3 - ext_4

            x_pos = item[5][0] + delta_x
            y_pos = item[5][1] + delta_y

            ax1.annotate('', (x_pos, y_pos), xytext=(item[5][0], item[5][1]),
                         fontsize=8, arrowprops={'arrowstyle': '->'}, color='black', zorder=3.)

            x_pos_text = item[5][0] + delta_x/2.
            y_pos_text = item[5][1] + delta_y/2.

            vector_len = math.sqrt(delta_x**2+delta_y**2)

            if item[3] == 'MgSiO3':
                dust_species = r'MgSiO$_{3}$'

            elif item[3] == 'Fe':
                dust_species = 'Fe'

            if item[4].is_integer():
                red_label = f'{dust_species} ({item[4]:.0f} µm)'

            else:
                red_label = f'{dust_species} ({item[4]:.1f} µm)'

            text = ax1.annotate(red_label, (x_pos_text, y_pos_text),
                                xytext=(-7.*delta_y/vector_len, 7.*delta_x/vector_len),
                                textcoords='offset points', fontsize=8., color='black',
                                ha='center', va='center')

            ax1.plot([item[5][0], x_pos], [item[5][1], y_pos], '-', color='white')

            sp1 = ax1.transData.transform_point((item[5][0], item[5][1]))
            sp2 = ax1.transData.transform_point((x_pos, y_pos))

            angle = np.degrees(np.arctan2(sp2[1]-sp1[1], sp2[0]-sp1[0]))
            text.set_rotation(angle)

    if objects is not None:
        for i, item in enumerate(objects):
            objdata = read_object.ReadObject(item[0])

            objphot1 = objdata.get_photometry(item[1][0])
            objphot2 = objdata.get_photometry(item[1][1])
            objphot3 = objdata.get_photometry(item[2][0])
            objphot4 = objdata.get_photometry(item[2][1])

            if objphot1.ndim == 2:
                print(f'Found {objphot1.shape[1]} values for filter {item[1][0]} of {item[0]}')
                print(f'so using the first value:  {objphot1[0, 0]} +/- {objphot1[1, 0]} mag')
                objphot1 = objphot1[:, 0]

            if objphot2.ndim == 2:
                print(f'Found {objphot2.shape[1]} values for filter {item[1][1]} of {item[0]}')
                print(f'so using the first value:  {objphot2[0, 0]} +/- {objphot2[1, 0]} mag')
                objphot2 = objphot2[:, 0]

            if objphot3.ndim == 2:
                print(f'Found {objphot3.shape[1]} values for filter {item[2][0]} of {item[0]}')
                print(f'so using the first value:  {objphot3[0, 0]} +/- {objphot3[1, 0]} mag')
                objphot3 = objphot3[:, 0]

            if objphot4.ndim == 2:
                print(f'Found {objphot4.shape[1]} values for filter {item[2][1]} of {item[0]}')
                print(f'so using the first value:  {objphot4[0, 0]} +/- {objphot4[1, 0]} mag')
                objphot4 = objphot4[:, 0]

            color1 = objphot1[0] - objphot2[0]
            color2 = objphot3[0] - objphot4[0]

            error1 = math.sqrt(objphot1[1]**2+objphot2[1]**2)
            error2 = math.sqrt(objphot3[1]**2+objphot4[1]**2)

            if len(item) > 3 and item[3] is not None:
                kwargs = item[3]

            else:
                kwargs = {'marker': '>',
                          'ms': 6.,
                          'color': 'black',
                          'mfc': 'white',
                          'mec': 'black',
                          'label': 'Direct imaging'}

            ax1.errorbar(color1, color2, xerr=error1, yerr=error2, zorder=3, **kwargs)

            if companion_labels:
                if len(item) > 3:
                    kwargs = item[4]

                else:
                    kwargs = {'ha': 'left',
                              'va': 'bottom',
                              'fontsize': 8.5,
                              'xytext': (5., 5.),
                              'color': 'black'}

                ax1.annotate(objdata.object_name, (color1, color2), zorder=3,
                             textcoords='offset points', **kwargs)

    print(f'Plotting color-color diagram: {output}...', end='', flush=True)

    handles, labels = ax1.get_legend_handles_labels()

    if legend is not None:
        handles, labels = ax1.get_legend_handles_labels()

        # prevent duplicates
        by_label = dict(zip(labels, handles))

        if handles:
            ax1.legend(by_label.values(), by_label.keys(), loc=legend, fontsize=8.5,
                       frameon=False, numpoints=1)

    plt.savefig(os.getcwd()+'/'+output, bbox_inches='tight')
    plt.clf()
    plt.close()

    print(' [DONE]')



"""
Module with reading functionalities for isochrones.
"""

import configparser
import os
import warnings

from typing import Optional, Tuple

import h5py
import numpy as np

from typeguard import typechecked
from scipy.interpolate import griddata

from species.core import box
from species.read import read_model


class ReadIsochrone:
    """
    Class for reading isochrone data from the database.
    """

    @typechecked
    def __init__(self,
                 tag: str) -> None:
        """
        Parameters
        ----------
        tag : str
            Database tag of the isochrone data.

        Returns
        -------
        NoneType
            None
        """

        self.tag = tag

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

    @typechecked
    def get_isochrone(self,
                      age: float,
                      masses: np.ndarray,
                      filters_color: Optional[Tuple[str, str]] = None,
                      filter_mag: Optional[str] = None) -> box.IsochroneBox:
        """
        Function for selecting an isochrone.

        Parameters
        ----------
        age : float
            Age (Myr) at which the isochrone data is interpolated.
        masses : np.ndarray
            Masses (Mjup) at which the isochrone data is interpolated.
        filters_color : tuple(str, str), None
            Filter names for the color as listed in the file with the isochrone data. Not selected
            if set to ``None`` or if only evolutionary tracks are available.
        filter_mag : str, None
            Filter name for the absolute magnitude as listed in the file with the isochrone data.
            Not selected if set to ``None`` or if only evolutionary tracks are available.

        Returns
        -------
        species.core.box.IsochroneBox
            Box with the isochrone.
        """

        age_points = np.repeat(age, masses.shape[0])  # (Myr)

        color = None
        mag_abs = None

        index_teff = 2
        index_logg = 4

        with h5py.File(self.database, 'r') as h5_file:
            model = h5_file[f'isochrones/{self.tag}/evolution'].attrs['model']
            evolution = np.asarray(h5_file[f'isochrones/{self.tag}/evolution'])

            if model == 'baraffe':
                filters = list(h5_file[f'isochrones/{self.tag}/filters'])
                magnitudes = np.asarray(h5_file[f'isochrones/{self.tag}/magnitudes'])

        if model == 'baraffe':
            if filters_color is not None:
                index_color_1 = filters.index(filters_color[0])
                index_color_2 = filters.index(filters_color[1])

            if filter_mag is not None:
                index_mag = filters.index(filter_mag)

            if filters_color is not None:
                mag_color_1 = griddata(points=evolution[:, 0:2],
                                       values=magnitudes[:, index_color_1],
                                       xi=np.stack((age_points, masses), axis=1),
                                       method='linear',
                                       fill_value='nan',
                                       rescale=False)

                mag_color_2 = griddata(points=evolution[:, 0:2],
                                       values=magnitudes[:, index_color_2],
                                       xi=np.stack((age_points, masses), axis=1),
                                       method='linear',
                                       fill_value='nan',
                                       rescale=False)

                color = mag_color_1-mag_color_2

            if filter_mag is not None:
                mag_abs = griddata(points=evolution[:, 0:2],
                                   values=magnitudes[:, index_mag],
                                   xi=np.stack((age_points, masses), axis=1),
                                   method='linear',
                                   fill_value='nan',
                                   rescale=False)

        teff = griddata(points=evolution[:, 0:2],
                        values=evolution[:, index_teff],
                        xi=np.stack((age_points, masses), axis=1),
                        method='linear',
                        fill_value='nan',
                        rescale=False)

        logg = griddata(points=evolution[:, 0:2],
                        values=evolution[:, index_logg],
                        xi=np.stack((age_points, masses), axis=1),
                        method='linear',
                        fill_value='nan',
                        rescale=False)

        return box.create_box(boxtype='isochrone',
                              model=self.tag,
                              filters_color=filters_color,
                              filter_mag=filter_mag,
                              color=color,
                              magnitude=mag_abs,
                              teff=teff,
                              logg=logg,
                              masses=masses)

    @typechecked
    def get_color_magnitude(self,
                            age: float,
                            masses: np.ndarray,
                            model: str,
                            filters_color: Tuple[str, str],
                            filter_mag: str) -> box.ColorMagBox:
        """
        Function for calculating color-magnitude combinations from a selected isochrone.

        Parameters
        ----------
        age : float
            Age (Myr) at which the isochrone data is interpolated.
        masses : np.ndarray
            Masses (Mjup) at which the isochrone data is interpolated.
        model : str
            Atmospheric model used to compute the synthetic photometry.
        filters_color : tuple(str, str)
            Filter names for the color as listed in the file with the isochrone data. The filter
            names should be provided in the format of the SVO Filter Profile Service.
        filter_mag : str
            Filter name for the absolute magnitude as listed in the file with the isochrone data.
            The value should be equal to one of the ``filters_color`` values.

        Returns
        -------
        species.core.box.ColorMagBox
            Box with the color-magnitude data.
        """

        isochrone = self.get_isochrone(age=age,
                                       masses=masses,
                                       filters_color=None,
                                       filter_mag=None)

        model1 = read_model.ReadModel(model=model, filter_name=filters_color[0])
        model2 = read_model.ReadModel(model=model, filter_name=filters_color[1])

        if model1.get_parameters() != ['teff', 'logg']:
            raise ValueError('Creating synthetic colors and magnitudes from isochrones is '
                             'currently only implemented for models with only Teff and log(g) '
                             'as free parameters. Please create an issue on the Github page if '
                             'additional functionalities are required.')

        mag1 = np.zeros(isochrone.masses.shape[0])
        mag2 = np.zeros(isochrone.masses.shape[0])

        for i, mass_item in enumerate(isochrone.masses):
            model_param = {'teff': isochrone.teff[i],
                           'logg': isochrone.logg[i],
                           'mass': mass_item,
                           'distance': 10.}

            if np.isnan(isochrone.teff[i]):
                mag1[i] = np.nan
                mag2[i] = np.nan

                warnings.warn(f'The value of Teff is NaN for the following isochrone sample: '
                              f'{model_param}. Setting the magnitudes to NaN.')

            else:
                for item_bounds in model1.get_bounds():
                    if model_param[item_bounds] <= model1.get_bounds()[item_bounds][0]:
                        mag1[i] = np.nan
                        mag2[i] = np.nan

                        warnings.warn(f'The value of {item_bounds} is {model_param[item_bounds]}, '
                                      f'which is below the lower bound of the model grid '
                                      f'({model1.get_bounds()[item_bounds][0]}). Setting the '
                                      f'magnitudes to NaN for the following isochrone sample: '
                                      f'{model_param}.')

                    elif model_param[item_bounds] >= model1.get_bounds()[item_bounds][1]:
                        mag1[i] = np.nan
                        mag2[i] = np.nan

                        warnings.warn(f'The value of {item_bounds} is {model_param[item_bounds]}, '
                                      f'which is above the upper bound of the model grid '
                                      f'({model1.get_bounds()[item_bounds][1]}). Setting the '
                                      f'magnitudes to NaN for the following isochrone sample: '
                                      f'{model_param}.')

                if not np.isnan(mag1[i]):
                    mag1[i], _ = model1.get_magnitude(model_param)
                    mag2[i], _ = model2.get_magnitude(model_param)

        if filter_mag == filters_color[0]:
            abs_mag = mag1

        elif filter_mag == filters_color[1]:
            abs_mag = mag2

        else:
            raise ValueError('The argument of filter_mag should be equal to one of the two filter '
                             'values of filters_color.')

        return box.create_box(boxtype='colormag',
                              library=model,
                              object_type='model',
                              filters_color=filters_color,
                              filter_mag=filter_mag,
                              color=mag1-mag2,
                              magnitude=abs_mag,
                              sptype=masses,
                              names=None)

    @typechecked
    def get_color_color(self,
                        age: float,
                        masses: np.ndarray,
                        model: str,
                        filters_colors: Tuple[Tuple[str, str],
                                              Tuple[str, str]]) -> box.ColorColorBox:
        """
        Function for calculating color-magnitude combinations from a selected isochrone.

        Parameters
        ----------
        age : float
            Age (Myr) at which the isochrone data is interpolated.
        masses : np.ndarray
            Masses (Mjup) at which the isochrone data is interpolated.
        model : str
            Atmospheric model used to compute the synthetic photometry.
        filters_colors : tuple(tuple(str, str), tuple(str, str))
            Filter names for the colors as listed in the file with the isochrone data. The filter
            names should be provided in the format of the SVO Filter Profile Service.

        Returns
        -------
        species.core.box.ColorColorBox
            Box with the color-color data.
        """

        isochrone = self.get_isochrone(age=age,
                                       masses=masses,
                                       filters_color=None,
                                       filter_mag=None)

        model1 = read_model.ReadModel(model=model, filter_name=filters_colors[0][0])
        model2 = read_model.ReadModel(model=model, filter_name=filters_colors[0][1])
        model3 = read_model.ReadModel(model=model, filter_name=filters_colors[1][0])
        model4 = read_model.ReadModel(model=model, filter_name=filters_colors[1][1])

        if model1.get_parameters() != ['teff', 'logg']:
            raise ValueError('Creating synthetic colors and magnitudes from isochrones is '
                             'currently only implemented for models with only Teff and log(g) '
                             'as free parameters. Please contact Tomas Stolker if additional '
                             'functionalities are required.')

        mag1 = np.zeros(isochrone.masses.shape[0])
        mag2 = np.zeros(isochrone.masses.shape[0])
        mag3 = np.zeros(isochrone.masses.shape[0])
        mag4 = np.zeros(isochrone.masses.shape[0])

        for i, mass_item in enumerate(isochrone.masses):
            model_param = {'teff': isochrone.teff[i],
                           'logg': isochrone.logg[i],
                           'mass': mass_item,
                           'distance': 10.}

            if np.isnan(isochrone.teff[i]):
                mag1[i] = np.nan
                mag2[i] = np.nan
                mag3[i] = np.nan
                mag4[i] = np.nan

                warnings.warn(f'The value of Teff is NaN for the following isochrone sample: '
                              f'{model_param}. Setting the magnitudes to NaN.')

            else:
                for item_bounds in model1.get_bounds():
                    if model_param[item_bounds] <= model1.get_bounds()[item_bounds][0]:
                        mag1[i] = np.nan
                        mag2[i] = np.nan
                        mag3[i] = np.nan
                        mag4[i] = np.nan

                        warnings.warn(f'The value of {item_bounds} is {model_param[item_bounds]}, '
                                      f'which is below the lower bound of the model grid '
                                      f'({model1.get_bounds()[item_bounds][0]}). Setting the '
                                      f'magnitudes to NaN for the following isochrone sample: '
                                      f'{model_param}.')

                    elif model_param[item_bounds] >= model1.get_bounds()[item_bounds][1]:
                        mag1[i] = np.nan
                        mag2[i] = np.nan
                        mag3[i] = np.nan
                        mag4[i] = np.nan

                        warnings.warn(f'The value of {item_bounds} is {model_param[item_bounds]}, '
                                      f'which is above the upper bound of the model grid '
                                      f'({model1.get_bounds()[item_bounds][1]}). Setting the '
                                      f'magnitudes to NaN for the following isochrone sample: '
                                      f'{model_param}.')

                if not np.isnan(mag1[i]) and not np.isnan(mag2[i]) and\
                        not np.isnan(mag3[i]) and not np.isnan(mag4[i]):
                    mag1[i], _ = model1.get_magnitude(model_param)
                    mag2[i], _ = model2.get_magnitude(model_param)
                    mag3[i], _ = model3.get_magnitude(model_param)
                    mag4[i], _ = model4.get_magnitude(model_param)

        return box.create_box(boxtype='colorcolor',
                              library=model,
                              object_type='model',
                              filters=filters_colors,
                              color1=mag1-mag2,
                              color2=mag3-mag4,
                              sptype=masses,
                              names=None)



import os
import shutil
import urllib.request

import pytest
import numpy as np

import species
from species.util import test_util


class TestPhotometry:

    def setup_class(self):
        self.limit = 1e-8

        url = 'http://irtfweb.ifa.hawaii.edu/~spex/IRTF_Spectral_Library/Data/plnt_Jupiter.txt'
        urllib.request.urlretrieve(url, 'plnt_Jupiter.txt')

    def teardown_class(self):
        os.remove('species_database.hdf5')
        os.remove('species_config.ini')
        os.remove('plnt_Jupiter.txt')
        shutil.rmtree('data/')

    def test_species_init(self):
        test_util.create_config('./')
        species.SpeciesInit()

    def test_synthetic_photometry(self):
        species.SyntheticPhotometry('MKO/NSFCam.J')

    def test_magnitude_to_flux(self):
        synphot = species.SyntheticPhotometry('MKO/NSFCam.J')
        flux, error = synphot.magnitude_to_flux(20., error=0.5)

        assert flux == pytest.approx(3.104545900342411e-17, rel=self.limit, abs=0.)
        assert error == pytest.approx(1.4807688130194138e-17, rel=self.limit, abs=0.)

    def test_flux_to_magnitude(self):
        synphot = species.SyntheticPhotometry('MKO/NSFCam.J')
        app_mag, abs_mag = synphot.flux_to_magnitude(1e-10, error=None, distance=(50., None))

        assert app_mag[0] == pytest.approx(3.729995213054507, rel=self.limit, abs=0.)
        assert abs_mag[0] == pytest.approx(0.23514519137441336, rel=self.limit, abs=0.)

    def test_spectrum_to_flux(self):
        jup_wavel, jup_flux, jup_err = np.loadtxt('plnt_Jupiter.txt', unpack=True)

        synphot = species.SyntheticPhotometry('MKO/NSFCam.J')

        phot_flux, phot_error = synphot.spectrum_to_flux(jup_wavel,
                                                         jup_flux,
                                                         error=jup_err,
                                                         threshold=None)

        assert phot_flux == pytest.approx(1.802998152236653e-09, rel=self.limit, abs=0.)

        # The error is estimated with Monte Carlo sampling
        assert phot_error == pytest.approx(8.8e-14, rel=0., abs=2e-14)

    def test_spectrum_to_flux_no_error(self):
        jup_wavel, jup_flux, _ = np.loadtxt('plnt_Jupiter.txt', unpack=True)

        synphot = species.SyntheticPhotometry('MKO/NSFCam.J')

        phot_flux, phot_error = synphot.spectrum_to_flux(jup_wavel,
                                                         jup_flux,
                                                         error=None,
                                                         threshold=None)

        assert phot_flux == pytest.approx(1.802998152236653e-09, rel=self.limit, abs=0.)
        assert phot_error is None

    def test_spectrum_to_flux_threshold(self):
        jup_wavel, jup_flux, _ = np.loadtxt('plnt_Jupiter.txt', unpack=True)

        synphot = species.SyntheticPhotometry('MKO/NSFCam.J')

        phot_flux, phot_error = synphot.spectrum_to_flux(jup_wavel,
                                                         jup_flux,
                                                         error=None,
                                                         threshold=0.05)

        assert phot_flux == pytest.approx(1.802998152236653e-09, rel=self.limit, abs=0.)
        assert phot_error is None

    def test_spectrum_to_flux_photon_detector(self):
        jup_wavel, jup_flux, jup_err = np.loadtxt('plnt_Jupiter.txt', unpack=True)

        synphot = species.SyntheticPhotometry('Keck/NIRC2.J')

        phot_flux, phot_error = synphot.spectrum_to_flux(jup_wavel,
                                                         jup_flux,
                                                         error=jup_err,
                                                         threshold=None)

        assert phot_flux == pytest.approx(1.8139884828774647e-09, rel=self.limit, abs=0.)

        # The error is estimated with Monte Carlo sampling
        assert phot_error == pytest.approx(8.4e-14, rel=0., abs=2e-14)

    def test_spectrum_to_magnitude(self):
        jup_wavel, jup_flux, jup_err = np.loadtxt('plnt_Jupiter.txt', unpack=True)

        synphot = species.SyntheticPhotometry('MKO/NSFCam.J')

        app_mag, abs_mag = synphot.spectrum_to_magnitude(jup_wavel,
                                                         jup_flux,
                                                         error=jup_err,
                                                         distance=(1., 0.01),
                                                         threshold=None)

        assert app_mag[0] == pytest.approx(0.5900070089410099, rel=self.limit, abs=0.)
        assert abs_mag[0] == pytest.approx(5.59000700894101, rel=self.limit, abs=0.)

        # The error is estimated with Monte Carlo sampling
        assert app_mag[1] == pytest.approx(5.368048545366946e-05, rel=0., abs=1e-5)
        assert abs_mag[1] == pytest.approx(0.021714790446227043, rel=0., abs=1e-2)



"""
Module with functionalities for reading and writing of data.
"""

import configparser
import os
import warnings

from typing import Dict, List, Optional, Tuple, Union

import emcee
import h5py
import numpy as np
import tqdm

from astropy.io import fits
from typeguard import typechecked

from species.analysis import photometry
from species.core import box
from species.data import ames_cond, ames_dusty, atmo, blackbody, btcond, btcond_feh, btnextgen, \
                         btsettl, btsettl_cifist, companions, drift_phoenix, dust, exo_rem, \
                         filters, irtf, isochrones, leggett, petitcode, spex, vega, vlm_plx
from species.read import read_calibration, read_filter, read_model, read_planck
from species.util import data_util, dust_util, read_util


class Database:
    """
    Class with reading and writing functionalities for the HDF5 database.
    """

    @typechecked
    def __init__(self) -> None:
        """
        Returns
        -------
        NoneType
            None
        """

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']
        self.input_path = config['species']['data_folder']

    @typechecked
    def list_content(self) -> None:
        """
        Returns
        -------
        NoneType
            None
        """

        print('Database content:')

        @typechecked
        def descend(h5_object: Union[h5py._hl.files.File,
                                     h5py._hl.group.Group,
                                     h5py._hl.dataset.Dataset],
                    seperator: str = '') -> None:
            """
            Function for descending into an HDF5 dataset and printing its content.

            Parameters
            ----------
            h5_object : h5py._hl.files.File, h5py._hl.group.Group, h5py._hl.dataset.Dataset
                The ``h5py`` object.
            separator : str
                Separator that is used between items.

            Returns
            -------
            NoneType
                None
            """

            if isinstance(h5_object, (h5py._hl.files.File, h5py._hl.group.Group)):
                for key in h5_object.keys():
                    print(seperator+'- '+key+': '+str(h5_object[key]))
                    descend(h5_object[key], seperator=seperator+'\t')

            elif isinstance(h5_object, h5py._hl.dataset.Dataset):
                for key in h5_object.attrs.keys():
                    print(seperator+'- '+key+': '+str(h5_object.attrs[key]))

        with h5py.File(self.database, 'r') as hdf_file:
            descend(hdf_file)

    @staticmethod
    @typechecked
    def list_companions() -> None:
        """
        Returns
        -------
        NoneType
            None
        """

        for planet_name, planet_dict in companions.get_data().items():
            distance = planet_dict['distance']
            app_mag = planet_dict['app_mag']

            print(f'Object name = {planet_name}')
            print(f'Distance (pc) = {distance[0]} +/- {distance[1]}')

            for mag_name, mag_dict in app_mag.items():
                print(f'{mag_name} (mag) = {mag_dict[0]} +/- {mag_dict[1]}')

            print()

    @typechecked
    def delete_data(self,
                    dataset: str) -> None:
        """
        Function for deleting a dataset from the HDF5 database.

        Parameters
        ----------
        dataset : str
            Dataset path in the HDF5 database.

        Returns
        -------
        NoneType
            None
        """

        with h5py.File(self.database, 'a') as hdf_file:
            if dataset in hdf_file:
                print(f'Deleting data: {dataset}...', end='', flush=True)
                del hdf_file[dataset]
                print(' [DONE]')

            else:
                warnings.warn(f'The dataset {dataset} is not found in {self.database}.')

    @typechecked
    def add_companion(self,
                      name: Union[Optional[str], Optional[List[str]]] = None) -> None:
        """
        Function for adding the magnitudes of directly imaged planets and brown dwarfs from
        :class:`~species.data.companions.get_data` to the database.

        Parameters
        ----------
        name : str, list(str), None
            Name or list with names of the directly imaged planets and brown dwarfs (e.g.
            ``'HR 8799 b'`` or ``['HR 8799 b', '51 Eri b', 'PZ Tel B']``). All the available
            companion data are added if set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        if isinstance(name, str):
            name = list((name, ))

        data = companions.get_data()

        if name is None:
            name = data.keys()

        for item in name:
            self.add_object(object_name=item,
                            distance=data[item]['distance'],
                            app_mag=data[item]['app_mag'])

    @typechecked
    def add_dust(self) -> None:
        """
        Function for adding optical constants of MgSiO3 and Fe, and MgSiO3 cross sections for
        a log-normal and power-law size distribution to the database. The optical constants have
        been compiled by Mollière et al. (2019) for petitRADTRANS from the following sources:

        - MgSiO3, crystalline
            - Scott & Duley (1996), ApJS, 105, 401
            - Jäger et al. (1998), A&A, 339, 904

        - MgSiO3, amorphous
            - Jäger et al. (2003), A&A, 408, 193

        - Fe, crystalline
            - Henning & Stognienko (1996), A&A, 311, 291

        - Fe, amorphous
            - Pollack et al. (1994), ApJ, 421, 615

        Returns
        -------
        NoneType
            None
        """

        h5_file = h5py.File(self.database, 'a')

        if 'dust' in h5_file:
            del h5_file['dust']

        h5_file.create_group('dust')

        dust.add_optical_constants(self.input_path, h5_file)
        dust.add_cross_sections(self.input_path, h5_file)

        h5_file.close()

    @typechecked
    def add_filter(self,
                   filter_name: str,
                   filename: Optional[str] = None,
                   detector_type: str = 'photon') -> None:
        """
        Function for adding a filter profile to the database, either from the SVO Filter profile
        Service or from an input file. Additional filters that are automatically added are
        Magellan/VisAO.rp, Magellan/VisAO.ip, Magellan/VisAO.zp, Magellan/VisAO.Ys, ALMA/band6,
        and ALMA/band7.

        Parameters
        ----------
        filter_name : str
            Filter name from the SVO Filter Profile Service (e.g., 'Paranal/NACO.Lp') or a
            user-defined name if a ``filename`` is specified.
        filename : str
            Filename of the filter profile. The first column should contain the wavelength
            (um) and the second column the fractional transmission. The profile is downloaded from
            the SVO Filter Profile Service if the argument of ``filename`` is set to ``None``.
        detector_type : str
            The detector type ('photon' or 'energy'). The argument is only used if a ``filename``
            is provided. Otherwise, for filters that are fetched from the SVO website, the detector
            type is read from the SVO data. The detector type determines if a wavelength factor
            is included in the integral for the synthetic photometry.

        Returns
        -------
        NoneType
            None
        """

        print(f'Adding filter: {filter_name}...', end='', flush=True)

        filter_split = filter_name.split('/')

        h5_file = h5py.File(self.database, 'a')

        if f'filters/{filter_name}' in h5_file:
            del h5_file[f'filters/{filter_name}']

        if 'filters' not in h5_file:
            h5_file.create_group('filters')

        if f'filters/{filter_split[0]}' not in h5_file:
            h5_file.create_group(f'filters/{filter_split[0]}')

        if filename is not None:
            data = np.loadtxt(filename)
            wavelength = data[:, 0]
            transmission = data[:, 1]

        else:
            wavelength, transmission, detector_type = filters.download_filter(filter_name)

        if wavelength is not None and transmission is not None:
            dset = h5_file.create_dataset(f'filters/{filter_name}',
                                          data=np.column_stack((wavelength, transmission)))

            dset.attrs['det_type'] = str(detector_type)

        h5_file.close()

        print(' [DONE]')

    @typechecked
    def add_isochrones(self,
                       filename: str,
                       tag: str,
                       model: str = 'baraffe') -> None:
        """
        Function for adding isochrone data to the database.

        Parameters
        ----------
        filename : str
            Filename with the isochrone data.
        tag : str
            Database tag name where the isochrone that will be stored.
        model : str
            Evolutionary model ('baraffe' or 'marleau'). For 'baraffe' models, the isochrone data
            can be downloaded from https://phoenix.ens-lyon.fr/Grids/. For 'marleau' models, the
            data can be requested from Gabriel Marleau.

        Returns
        -------
        NoneType
            None
        """

        h5_file = h5py.File(self.database, 'a')

        if 'isochrones' not in h5_file:
            h5_file.create_group('isochrones')

        if 'isochrones/'+tag in h5_file:
            del h5_file[f'isochrones/{tag}']

        if model[0:7] == 'baraffe':
            isochrones.add_baraffe(h5_file, tag, filename)

        elif model[0:7] == 'marleau':
            isochrones.add_marleau(h5_file, tag, filename)

        h5_file.close()

    @typechecked
    def add_model(self,
                  model: str,
                  wavel_range: Optional[Tuple[float, float]] = None,
                  spec_res: Optional[float] = None,
                  teff_range: Optional[Tuple[float, float]] = None,
                  data_folder: Optional[str] = None) -> None:
        """
        Parameters
        ----------
        model : str
            Model name ('ames-cond', 'ames-dusty', 'atmo', 'bt-settl', 'bt-settl-cifist',
            'bt-nextgen', 'drift-phoenix', 'petitcode-cool-clear', 'petitcode-cool-cloudy',
            'petitcode-hot-clear', 'petitcode-hot-cloudy', 'exo-rem', 'blackbody', bt-cond',
            or 'bt-cond-feh).
        wavel_range : tuple(float, float), None
            Wavelength range (um). Optional for the DRIFT-PHOENIX and petitCODE models. For
            these models, the original wavelength points are used if set to ``None``.
            which case the argument can be set to ``None``.
        spec_res : float, None
            Spectral resolution. The parameter is optional for the DRIFT-PHOENIX, petitCODE,
            BT-Settl, and Exo-REM models. The argument is only used if ``wavel_range`` is not
            ``None``.
        teff_range : tuple(float, float), None
            Effective temperature range (K). Setting the value to None for will add all available
            temperatures.
        data_folder : str, None
            Folder with input data. This parameter has been deprecated since all model spectra
            are publicly available.

        Returns
        -------
        NoneType
            None
        """

        if data_folder is not None:
            warnings.warn('The \'data_folder\' parameter has been deprecated since '
                          'all supported model spectra are publicly available. The'
                          'parameter will therefore be ignored and will cause an error '
                          'in a future release.')

        # proprietary = ['petitcode-hot-clear', 'petitcode-hot-cloudy']

        # if model in proprietary and data_folder is None:
        #     raise ValueError(f'The {model} model is not publicly available and needs to '
        #                      f'be imported by setting the \'data_folder\' parameter.')

        # if model in ['bt-nextgen'] and wavel_range is None:
        #     raise ValueError(f'The \'wavel_range\' should be set for the \'{model}\' models to '
        #                      f'resample the original spectra on a fixed wavelength grid.')

        # if model in ['bt-nextgen'] and spec_res is None:
        #     raise ValueError(f'The \'spec_res\' should be set for the \'{model}\' models to '
        #                      f'resample the original spectra on a fixed wavelength grid.')

        # if model == 'bt-nextgen' and teff_range is None:
        #     warnings.warn('The temperature range is not restricted with the \'teff_range\' '
        #                   'parameter. Therefore, adding the BT-Settl or BT-NextGen spectra '
        #                   'will be very slow.')

        h5_file = h5py.File(self.database, 'a')

        if 'models' not in h5_file:
            h5_file.create_group('models')

        if model == 'ames-cond':
            ames_cond.add_ames_cond(self.input_path,
                                    h5_file,
                                    wavel_range,
                                    teff_range,
                                    spec_res)

            data_util.add_missing(model, ['teff', 'logg'], h5_file)

        elif model == 'ames-dusty':
            ames_dusty.add_ames_dusty(self.input_path,
                                      h5_file,
                                      wavel_range,
                                      teff_range,
                                      spec_res)

            data_util.add_missing(model, ['teff', 'logg'], h5_file)

        elif model == 'atmo':
            atmo.add_atmo(self.input_path,
                          h5_file,
                          wavel_range,
                          teff_range,
                          spec_res)

            data_util.add_missing(model, ['teff', 'logg'], h5_file)

        elif model == 'blackbody':
            blackbody.add_blackbody(self.input_path,
                                    h5_file,
                                    wavel_range,
                                    teff_range,
                                    spec_res)

            data_util.add_missing(model, ['teff'], h5_file)

        elif model == 'bt-cond':
            btcond.add_btcond(self.input_path,
                              h5_file,
                              wavel_range,
                              teff_range,
                              spec_res)

            data_util.add_missing(model, ['teff', 'logg'], h5_file)

        elif model == 'bt-cond-feh':
            btcond_feh.add_btcond_feh(self.input_path,
                                      h5_file,
                                      wavel_range,
                                      teff_range,
                                      spec_res)

            data_util.add_missing(model, ['teff', 'logg', 'feh'], h5_file)

        elif model == 'bt-settl':
            btsettl.add_btsettl(self.input_path,
                                h5_file,
                                wavel_range,
                                teff_range,
                                spec_res)

            data_util.add_missing(model, ['teff', 'logg'], h5_file)

        elif model == 'bt-settl-cifist':
            btsettl_cifist.add_btsettl(self.input_path,
                                       h5_file,
                                       wavel_range,
                                       teff_range,
                                       spec_res)

            data_util.add_missing(model, ['teff', 'logg'], h5_file)

        elif model == 'bt-nextgen':
            btnextgen.add_btnextgen(self.input_path,
                                    h5_file,
                                    wavel_range,
                                    teff_range,
                                    spec_res)

            data_util.add_missing(model, ['teff', 'logg', 'feh'], h5_file)

        elif model == 'drift-phoenix':
            drift_phoenix.add_drift_phoenix(self.input_path,
                                            h5_file,
                                            wavel_range,
                                            teff_range,
                                            spec_res)

            data_util.add_missing(model, ['teff', 'logg', 'feh'], h5_file)

        elif model == 'petitcode-cool-clear':
            petitcode.add_petitcode_cool_clear(self.input_path,
                                               h5_file,
                                               wavel_range,
                                               teff_range,
                                               spec_res)

            data_util.add_missing(model, ['teff', 'logg', 'feh'], h5_file)

        elif model == 'petitcode-cool-cloudy':
            petitcode.add_petitcode_cool_cloudy(self.input_path,
                                                h5_file,
                                                wavel_range,
                                                teff_range,
                                                spec_res)

            data_util.add_missing(model, ['teff', 'logg', 'feh', 'fsed'], h5_file)

        elif model == 'petitcode-hot-clear':
            petitcode.add_petitcode_hot_clear(self.input_path,
                                              h5_file,
                                              wavel_range,
                                              teff_range,
                                              spec_res)

            data_util.add_missing(model, ['teff', 'logg', 'feh', 'co'], h5_file)

        elif model == 'petitcode-hot-cloudy':
            petitcode.add_petitcode_hot_cloudy(self.input_path,
                                               h5_file,
                                               wavel_range,
                                               teff_range,
                                               spec_res)

            data_util.add_missing(model, ['teff', 'logg', 'feh', 'co', 'fsed'], h5_file)

        elif model == 'exo-rem':
            exo_rem.add_exo_rem(self.input_path,
                                h5_file,
                                wavel_range,
                                teff_range,
                                spec_res)

            data_util.add_missing(model, ['teff', 'logg', 'feh', 'co'], h5_file)

        else:
            raise ValueError(f'The {model} atmospheric model is not available. Please choose from '
                             f'\'ames-cond\', \'ames-dusty\', \'atmo\', \'bt-settl\', '
                             f'\'bt-nextgen\', \'drift-phoexnix\', \'petitcode-cool-clear\', '
                             f'\'petitcode-cool-cloudy\', \'petitcode-hot-clear\', '
                             f'\'petitcode-hot-cloudy\', \'exo-rem\', \'bt-settl-cifist\', '
                             f'\'bt-cond\', \'bt-cond-feh\', \'blackbody\'.')

        h5_file.close()

    @typechecked
    def add_object(self,
                   object_name: str,
                   distance: Optional[Tuple[float, float]] = None,
                   app_mag: Optional[Dict[str,
                                          Union[Tuple[float, float],
                                                List[Tuple[float, float]]]]] = None,
                   flux_density: Optional[Dict[str, Tuple[float, float]]] = None,
                   spectrum: Optional[Dict[str,
                                           Tuple[str,
                                                 Optional[str],
                                                 Optional[float]]]] = None,
                   deredden: Union[Dict[str, float], float] = None) -> None:
        """
        Function for adding the photometric and/or spectroscopic data of an object to the database.

        Parameters
        ----------
        object_name: str
            Object name that will be used as label in the database.
        distance : tuple(float, float), None
            Distance and uncertainty (pc). Not stored if set to None.
        app_mag : dict, None
            Dictionary with the filter names, apparent magnitudes, and uncertainties. For example,
            ``{'Paranal/NACO.Lp': (15., 0.2), 'Paranal/NACO.Mp': (13., 0.3)}``. For the use of
            duplicate filter names, the magnitudes have to be provided in a list, for example
            ``{'Paranal/NACO.Lp': [(15., 0.2), (14.5, 0.5)], 'Paranal/NACO.Mp': (13., 0.3)}``.
            No photometric data is stored if set to ``None``.
        flux_density : dict, None
            Dictionary with filter names, flux densities (W m-2 um-1), and uncertainties
            (W m-1 um-1). For example, ``{'Paranal/NACO.Lp': (1e-15, 1e-16)}``. Currently,
            the use of duplicate filters is not implemented. The use of ``app_mag`` is preferred
            over ``flux_density`` because with ``flux_density`` only fluxes are stored while with
            ``app_mag`` both magnitudes and fluxes. However, ``flux_density`` can be used in case
            the magnitudes and/or filter profiles are not available. In that case, the fluxes can
            still be selected with ``inc_phot`` in :class:`~species.analysis.fit_model.FitModel`.
            The argument of ``flux_density`` is ignored if set to ``None``.
        spectrum : dict, None
            Dictionary with the spectrum, optional covariance matrix, and spectral resolution for
            each instrument. The input data can either have a FITS or ASCII format. The spectra
            should have 3 columns with wavelength (um), flux (W m-2 um-1), and uncertainty
            (W m-2 um-1). The covariance matrix should be 2D with the same number of wavelength
            points as the spectrum. For example, ``{'SPHERE': ('spectrum.dat', 'covariance.fits',
            50.)}``. No covariance data is stored if set to None, for example, ``{'SPHERE':
            ('spectrum.dat', None, 50.)}``. The ``spectrum`` parameter is ignored if set to None.
            For GRAVITY data, the same FITS file can be provided as spectrum and covariance matrix.
        deredden : dict, float, None
            Dictionary with ``spectrum`` and ``app_mag`` names that will de dereddened with the
            provided A_V. For example, ``deredden={'SPHERE': 1.5, 'Keck/NIRC2.J': 1.5}`` will
            deredden the provided spectrum named 'SPHERE' and the Keck/NIRC2 J-band photometry with
            a visual extinction of 1.5. For photometric fluxes, the filter-averaged extinction is
            used for the dereddening.

        Returns
        -------
        NoneType
            None
        """

        h5_file = h5py.File(self.database, 'a')

        if deredden is None:
            deredden = {}

        if app_mag is not None:
            if 'spectra/calibration/vega' not in h5_file:
                self.add_spectrum('vega')

            for item in app_mag:
                if f'filters/{item}' not in h5_file:
                    self.add_filter(item)

        if flux_density is not None:
            if 'spectra/calibration/vega' not in h5_file:
                self.add_spectrum('vega')

            for item in flux_density:
                if f'filters/{item}' not in h5_file:
                    self.add_filter(item)

        print(f'Adding object: {object_name}')

        if 'objects' not in h5_file:
            h5_file.create_group('objects')

        if f'objects/{object_name}' not in h5_file:
            h5_file.create_group(f'objects/{object_name}')

        if distance is not None:
            print(f'   - Distance (pc) = {distance[0]:.2f} +/- {distance[1]:.2f}')

            if f'objects/{object_name}/distance' in h5_file:
                del h5_file[f'objects/{object_name}/distance']

            h5_file.create_dataset(f'objects/{object_name}/distance',
                                   data=distance)  # (pc)

        flux = {}
        error = {}
        dered_phot = {}

        if app_mag is not None:
            for mag_item in app_mag:
                if isinstance(deredden, float) or mag_item in deredden:
                    read_filt = read_filter.ReadFilter(mag_item)
                    filter_profile = read_filt.get_filter()

                    if isinstance(deredden, float):
                        ext_mag = dust_util.ism_extinction(deredden, 3.1, filter_profile[:, 0])

                    else:
                        ext_mag = dust_util.ism_extinction(deredden[mag_item], 3.1,
                                                           filter_profile[:, 0])

                    synphot = photometry.SyntheticPhotometry(mag_item)

                    dered_phot[mag_item], _ = synphot.spectrum_to_flux(filter_profile[:, 0],
                                                                       10.**(0.4*ext_mag))

                else:
                    dered_phot[mag_item] = 1.

                if isinstance(app_mag[mag_item], tuple):

                    try:
                        synphot = photometry.SyntheticPhotometry(mag_item)

                        flux[mag_item], error[mag_item] = synphot.magnitude_to_flux(
                            app_mag[mag_item][0], app_mag[mag_item][1])

                        flux[mag_item] *= dered_phot[mag_item]

                    except KeyError:
                        warnings.warn(f'Filter \'{mag_item}\' is not available on the SVO Filter '
                                      f'Profile Service so a flux calibration can not be done. '
                                      f'Please add the filter manually with the \'add_filter\' '
                                      f'function. For now, only the \'{mag_item}\' magnitude of '
                                      f'\'{object_name}\' is stored.')

                        # Write NaNs if the filter is not available
                        flux[mag_item], error[mag_item] = np.nan, np.nan

                elif isinstance(app_mag[mag_item], list):
                    flux_list = []
                    error_list = []

                    for i, dupl_item in enumerate(app_mag[mag_item]):

                        try:
                            synphot = photometry.SyntheticPhotometry(mag_item)

                            flux_dupl, error_dupl = synphot.magnitude_to_flux(
                                dupl_item[0], dupl_item[1])

                            flux_dupl *= dered_phot[mag_item]

                        except KeyError:
                            warnings.warn(f'Filter \'{mag_item}\' is not available on the SVO '
                                          f'Filter Profile Service so a flux calibration can not '
                                          f'be done. Please add the filter manually with the '
                                          f'\'add_filter\' function. For now, only the '
                                          f'\'{mag_item}\' magnitude of \'{object_name}\' is '
                                          f'stored.')

                            # Write NaNs if the filter is not available
                            flux_dupl, error_dupl = np.nan, np.nan

                        flux_list.append(flux_dupl)
                        error_list.append(error_dupl)

                    flux[mag_item] = flux_list
                    error[mag_item] = error_list

                else:
                    raise ValueError('The values in the dictionary with magnitudes should be '
                                     'tuples or a list with tuples (in case duplicate filter '
                                     'names are required).')

            for mag_item in app_mag:
                if f'objects/{object_name}/{mag_item}' in h5_file:
                    del h5_file[f'objects/{object_name}/{mag_item}']

                if isinstance(app_mag[mag_item], tuple):
                    n_phot = 1

                    app_mag[mag_item] = (app_mag[mag_item][0] - 2.5*np.log10(dered_phot[mag_item]),
                                         app_mag[mag_item][1])

                    print(f'   - {mag_item}:')

                    print(f'      - Apparent magnitude = {app_mag[mag_item][0]:.2f} +/- '
                          f'{app_mag[mag_item][1]:.2f}')

                    print(f'      - Flux (W m-2 um-1) = {flux[mag_item]:.2e} +/- '
                          f'{error[mag_item]:.2e}')

                    if isinstance(deredden, float):
                        print(f'      - Dereddening A_V: {deredden}')

                    elif mag_item in deredden:
                        print(f'      - Dereddening A_V: {deredden[mag_item]}')

                    data = np.asarray([app_mag[mag_item][0],
                                       app_mag[mag_item][1],
                                       flux[mag_item],
                                       error[mag_item]])

                elif isinstance(app_mag[mag_item], list):
                    n_phot = len(app_mag[mag_item])
                    print(f'   - {mag_item} ({n_phot} values):')

                    mag_list = []
                    mag_err_list = []

                    for i, dupl_item in enumerate(app_mag[mag_item]):
                        dered_mag = app_mag[mag_item][i][0] - 2.5*np.log10(dered_phot[mag_item])
                        app_mag_item = (dered_mag, app_mag[mag_item][i][1])

                        print(f'      - Apparent magnitude = {app_mag_item[0]:.2f} +/- '
                              f'{app_mag_item[1]:.2f}')

                        print(f'      - Flux (W m-2 um-1) = {flux[mag_item][i]:.2e} +/- '
                              f'{error[mag_item][i]:.2e}')

                        mag_list.append(app_mag_item[0])
                        mag_err_list.append(app_mag_item[1])

                        if isinstance(deredden, float):
                            print(f'      - Dereddening A_V: {deredden}')

                        elif mag_item in deredden:
                            print(f'      - Dereddening A_V: {deredden[mag_item]}')

                    data = np.asarray([mag_list,
                                       mag_err_list,
                                       flux[mag_item],
                                       error[mag_item]])

                # (mag), (mag), (W m-2 um-1), (W m-2 um-1)
                dset = h5_file.create_dataset(f'objects/{object_name}/{mag_item}',
                                              data=data)

                dset.attrs['n_phot'] = n_phot

        if flux_density is not None:
            for flux_item in flux_density:
                if isinstance(deredden, float) or flux_item in deredden:
                    warnings.warn(f'The deredden parameter is not supported by flux_density. '
                                  f'Please use app_mag instead or contact Tomas Stolker and ask '
                                  f'if the flux_density support can be added. Ignoring the '
                                  f'dereddening of {flux_item}.')

                if f'objects/{object_name}/{flux_item}' in h5_file:
                    del h5_file[f'objects/{object_name}/{flux_item}']

                if isinstance(flux_density[flux_item], tuple):
                    print(f'   - {flux_item}:')

                    print(f'      - Flux (W m-2 um-1) = {flux_density[flux_item][0]:.2e} +/- '
                          f'{flux_density[flux_item][1]:.2e}')

                    data = np.asarray([np.nan,
                                       np.nan,
                                       flux_density[flux_item][0],
                                       flux_density[flux_item][1]])

                    # None, None, (W m-2 um-1), (W m-2 um-1)
                    dset = h5_file.create_dataset(f'objects/{object_name}/{flux_item}',
                                                  data=data)

                    dset.attrs['n_phot'] = n_phot

        if spectrum is not None:
            read_spec = {}
            read_cov = {}

            if f'objects/{object_name}/spectrum' in h5_file:
                del h5_file[f'objects/{object_name}/spectrum']

            # Read spectra

            for key, value in spectrum.items():
                if value[0].endswith('.fits'):
                    with fits.open(value[0]) as hdulist:
                        if 'INSTRU' in hdulist[0].header and \
                                hdulist[0].header['INSTRU'] == 'GRAVITY':
                            # Read data from a FITS file with the GRAVITY format
                            print('   - GRAVITY spectrum:')

                            gravity_object = hdulist[0].header['OBJECT']
                            print(f'      - Object: {gravity_object}')

                            wavelength = hdulist[1].data['WAVELENGTH']  # (um)
                            flux = hdulist[1].data['FLUX']  # (W m-2 um-1)
                            covariance = hdulist[1].data['COVARIANCE']  # (W m-2 um-1)^2
                            error = np.sqrt(np.diag(covariance))  # (W m-2 um-1)

                            read_spec[key] = np.column_stack([wavelength, flux, error])

                        else:
                            # Otherwise try to read a 2D dataset with 3 columns
                            print('   - Spectrum:')

                            for i, hdu_item in enumerate(hdulist):
                                data = np.asarray(hdu_item.data)

                                if data.ndim == 2 and 3 in data.shape and key not in read_spec:
                                    read_spec[key] = data

                            if key not in read_spec:
                                raise ValueError(f'The spectrum data from {value[0]} can not be '
                                                 f'read. The data format should be 2D with 3 '
                                                 f'columns.')

                else:
                    try:
                        data = np.loadtxt(value[0])
                    except UnicodeDecodeError:
                        raise ValueError(f'The spectrum data from {value[0]} can not be read. '
                                         f'Please provide a FITS or ASCII file.')

                    if data.ndim != 2 or 3 not in data.shape:
                        raise ValueError(f'The spectrum data from {value[0]} can not be read. The '
                                         f'data format should be 2D with 3 columns.')

                    print('   - Spectrum:')
                    read_spec[key] = data

                if isinstance(deredden, float):
                    ext_mag = dust_util.ism_extinction(deredden, 3.1, read_spec[key][:, 0])
                    read_spec[key][:, 1] *= 10.**(0.4*ext_mag)

                elif key in deredden:
                    ext_mag = dust_util.ism_extinction(deredden[key], 3.1, read_spec[key][:, 0])
                    read_spec[key][:, 1] *= 10.**(0.4*ext_mag)

                wavelength = read_spec[key][:, 0]
                flux = read_spec[key][:, 1]
                error = read_spec[key][:, 2]

                print(f'      - Database tag: {key}')
                print(f'      - Filename: {value[0]}')
                print(f'      - Data shape: {read_spec[key].shape}')
                print(f'      - Wavelength range (um): {wavelength[0]:.2f} - {wavelength[-1]:.2f}')
                print(f'      - Mean flux (W m-2 um-1): {np.mean(flux):.2e}')
                print(f'      - Mean error (W m-2 um-1): {np.mean(error):.2e}')

                if isinstance(deredden, float):
                    print(f'      - Dereddening A_V: {deredden}')

                elif key in deredden:
                    print(f'      - Dereddening A_V: {deredden[key]}')

            # Read covariance matrix

            for key, value in spectrum.items():
                if value[1] is None:
                    read_cov[key] = None

                elif value[1].endswith('.fits'):
                    with fits.open(value[1]) as hdulist:
                        if 'INSTRU' in hdulist[0].header and \
                                hdulist[0].header['INSTRU'] == 'GRAVITY':
                            # Read data from a FITS file with the GRAVITY format
                            print('   - GRAVITY covariance matrix:')

                            gravity_object = hdulist[0].header['OBJECT']
                            print(f'      - Object: {gravity_object}')

                            read_cov[key] = hdulist[1].data['COVARIANCE']  # (W m-2 um-1)^2

                        else:
                            # Otherwise try to read a square, 2D dataset
                            print('   - Covariance matrix:')

                            for i, hdu_item in enumerate(hdulist):
                                data = np.asarray(hdu_item.data)

                                corr_warn = f'The covariance matrix from {value[1]} contains ' \
                                            f'ones along the diagonal. Converting this ' \
                                            f'correlation  matrix into a covariance matrix.'

                                if data.ndim == 2 and data.shape[0] == data.shape[1]:
                                    if key not in read_cov:
                                        if data.shape[0] == read_spec[key].shape[0]:
                                            if np.all(np.diag(data) == 1.):
                                                warnings.warn(corr_warn)

                                                read_cov[key] = data_util.correlation_to_covariance(
                                                    data, read_spec[key][:, 2])

                                            else:
                                                read_cov[key] = data

                            if key not in read_cov:
                                raise ValueError(f'The covariance matrix from {value[1]} can not '
                                                 f'be read. The data format should be 2D with the '
                                                 f'same number of wavelength points as the '
                                                 f'spectrum.')

                else:
                    try:
                        data = np.loadtxt(value[1])
                    except UnicodeDecodeError:
                        raise ValueError(f'The covariance matrix from {value[1]} can not be read. '
                                         f'Please provide a FITS or ASCII file.')

                    if data.ndim != 2 or 3 not in data.shape:
                        raise ValueError(f'The covariance matrix from {value[1]} can not be read. '
                                         f'The data format should be 2D with the same number of '
                                         f'wavelength points as the spectrum.')

                    print('   - Covariance matrix:')

                    if np.all(np.diag(data) == 1.):
                        warnings.warn(f'The covariance matrix from {value[1]} contains ones on '
                                      f'the diagonal. Converting this correlation matrix into a '
                                      f'covariance matrix.')

                        read_cov[key] = data_util.correlation_to_covariance(
                            data, read_spec[key][:, 2])

                    else:
                        read_cov[key] = data

                if read_cov[key] is not None:
                    print(f'      - Database tag: {key}')
                    print(f'      - Filename: {value[1]}')
                    print(f'      - Data shape: {read_cov[key].shape}')

            print('   - Spectral resolution:')

            for key, value in spectrum.items():

                h5_file.create_dataset(f'objects/{object_name}/spectrum/{key}/spectrum',
                                       data=read_spec[key])

                if read_cov[key] is not None:
                    h5_file.create_dataset(f'objects/{object_name}/spectrum/{key}/covariance',
                                           data=read_cov[key])

                    h5_file.create_dataset(f'objects/{object_name}/spectrum/{key}/inv_covariance',
                                           data=np.linalg.inv(read_cov[key]))

                dset = h5_file[f'objects/{object_name}/spectrum/{key}']

                if value[2] is None:
                    print(f'      - {key}: None')
                    dset.attrs['specres'] = 0.

                else:
                    print(f'      - {key}: {value[2]:.1f}')
                    dset.attrs['specres'] = value[2]

        h5_file.close()

    @typechecked
    def add_photometry(self,
                       phot_library: str) -> None:
        """
        Parameters
        ----------
        phot_library : str
            Photometric library ('vlm-plx' or 'leggett').

        Returns
        -------
        NoneType
            None
        """

        h5_file = h5py.File(self.database, 'a')

        if 'photometry' not in h5_file:
            h5_file.create_group('photometry')

        if 'photometry/'+phot_library in h5_file:
            del h5_file['photometry/'+phot_library]

        if phot_library[0:7] == 'vlm-plx':
            vlm_plx.add_vlm_plx(self.input_path, h5_file)

        elif phot_library[0:7] == 'leggett':
            leggett.add_leggett(self.input_path, h5_file)

        h5_file.close()

    @typechecked
    def add_calibration(self,
                        tag: str,
                        filename: Optional[str] = None,
                        data: Optional[np.ndarray] = None,
                        units: Optional[Dict[str, str]] = None,
                        scaling: Optional[Tuple[float, float]] = None) -> None:
        """
        Function for adding a calibration spectrum to the database.

        Parameters
        ----------
        tag : str
            Tag name in the database.
        filename : str, None
            Filename with the calibration spectrum. The first column should contain the wavelength
            (um), the second column the flux density (W m-2 um-1), and the third column
            the error (W m-2 um-1). The ``data`` argument is used if set to ``None``.
        data : np.ndarray, None
            Spectrum stored as 3D array with shape ``(n_wavelength, 3)``. The first column should
            contain the wavelength (um), the second column the flux density (W m-2 um-1),
            and the third column the error (W m-2 um-1). The ``filename`` argument is used if set
            to ``None``.
        units : dict, None
            Dictionary with the wavelength and flux units, e.g. ``{'wavelength': 'angstrom',
            'flux': 'w m-2'}``. The default units (um and W m-2 um-1) are used if set to ``None``.
        scaling : tuple(float, float), None
            Scaling for the wavelength and flux as ``(scaling_wavelength, scaling_flux)``. Not used
            if set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        if filename is None and data is None:
            raise ValueError('Either the \'filename\' or \'data\' argument should be provided.')

        if scaling is None:
            scaling = (1., 1.)

        h5_file = h5py.File(self.database, 'a')

        if 'spectra/calibration' not in h5_file:
            h5_file.create_group('spectra/calibration')

        if 'spectra/calibration/'+tag in h5_file:
            del h5_file['spectra/calibration/'+tag]

        if filename is not None:
            data = np.loadtxt(filename)

        if units is None:
            wavelength = scaling[0]*data[:, 0]  # (um)
            flux = scaling[1]*data[:, 1]  # (W m-2 um-1)

        else:
            if units['wavelength'] == 'um':
                wavelength = scaling[0]*data[:, 0]  # (um)

            elif units['wavelength'] == 'angstrom':
                wavelength = scaling[0]*data[:, 0]*1e-4  # (um)

            if units['flux'] == 'w m-2 um-1':
                flux = scaling[1]*data[:, 1]  # (W m-2 um-1)

            elif units['flux'] == 'w m-2':
                if units['wavelength'] == 'um':
                    flux = scaling[1]*data[:, 1]/wavelength  # (W m-2 um-1)

        if data.shape[1] == 3:
            if units is None:
                error = scaling[1]*data[:, 2]  # (W m-2 um-1)

            else:
                if units['flux'] == 'w m-2 um-1':
                    error = scaling[1]*data[:, 2]  # (W m-2 um-1)

                elif units['flux'] == 'w m-2':
                    if units['wavelength'] == 'um':
                        error = scaling[1]*data[:, 2]/wavelength  # (W m-2 um-1)

        else:
            error = np.repeat(0., wavelength.size)

        print(f'Adding calibration spectrum: {tag}...', end='', flush=True)

        h5_file.create_dataset(f'spectra/calibration/{tag}',
                               data=np.vstack((wavelength, flux, error)))

        h5_file.close()

        print(' [DONE]')

    @typechecked
    def add_spectrum(self,
                     spec_library: str,
                     sptypes: Optional[List[str]] = None) -> None:
        """
        Parameters
        ----------
        spec_library : str
            Spectral library ('irtf' or 'spex').
        sptypes : list(str)
            Spectral types ('F', 'G', 'K', 'M', 'L', 'T'). Currently only implemented for 'irtf'.

        Returns
        -------
        NoneType
            None
        """

        h5_file = h5py.File(self.database, 'a')

        if 'spectra' not in h5_file:
            h5_file.create_group('spectra')

        if 'spectra/'+spec_library in h5_file:
            del h5_file['spectra/'+spec_library]

        if spec_library[0:5] == 'vega':
            vega.add_vega(self.input_path, h5_file)

        elif spec_library[0:5] == 'irtf':
            irtf.add_irtf(self.input_path, h5_file, sptypes)

        elif spec_library[0:5] == 'spex':
            spex.add_spex(self.input_path, h5_file)

        h5_file.close()

    @typechecked
    def add_samples(self,
                    sampler: str,
                    samples: np.ndarray,
                    ln_prob: np.ndarray,
                    mean_accept: Optional[float],
                    spectrum: Tuple[str, str],
                    tag: str,
                    modelpar: List[str],
                    distance: Optional[float],
                    spec_labels: Optional[List[str]]):
        """
        Parameters
        ----------
        sampler : str
            Sampler ('emcee' or 'multinest').
        samples : np.ndarray
            Samples of the posterior.
        ln_prob : np.ndarray
            Log posterior for each sample.
        mean_accept : float, None
            Mean acceptance fraction. Set to ``None`` when ``sampler`` is 'mulitinest'.
        spectrum : tuple(str, str)
            Tuple with the spectrum type ('model' or 'calibration') and spectrum name (e.g.
            'drift-phoenix').
        tag : str
            Database tag.
        modelpar : list(str)
            List with the model parameter names.
        distance : float, None
            Distance to the object (pc). Not used if set to ``None``.
        spec_labels : list(str), None
            List with the spectrum labels that are used for fitting an additional scaling
            parameter. Not used if set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        if spec_labels is None:
            spec_labels = []

        h5_file = h5py.File(self.database, 'a')

        if 'results' not in h5_file:
            h5_file.create_group('results')

        if 'results/fit' not in h5_file:
            h5_file.create_group('results/fit')

        if f'results/fit/{tag}' in h5_file:
            del h5_file[f'results/fit/{tag}']

        dset = h5_file.create_dataset(f'results/fit/{tag}/samples', data=samples)
        h5_file.create_dataset(f'results/fit/{tag}/ln_prob', data=ln_prob)

        dset.attrs['type'] = str(spectrum[0])
        dset.attrs['spectrum'] = str(spectrum[1])
        dset.attrs['n_param'] = int(len(modelpar))
        dset.attrs['sampler'] = str(sampler)

        if mean_accept is not None:
            dset.attrs['mean_accept'] = float(mean_accept)
            print(f'Mean acceptance fraction: {mean_accept:.3f}')

        if distance is not None:
            dset.attrs['distance'] = float(distance)

        count_scaling = 0

        for i, item in enumerate(modelpar):
            dset.attrs[f'parameter{i}'] = str(item)

            if item in spec_labels:
                dset.attrs[f'scaling{count_scaling}'] = str(item)
                count_scaling += 1

        dset.attrs['n_scaling'] = int(count_scaling)

        try:
            int_auto = emcee.autocorr.integrated_time(samples)
            print(f'Integrated autocorrelation time = {int_auto}')

        except emcee.autocorr.AutocorrError:
            int_auto = None

            print('The chain is shorter than 50 times the integrated autocorrelation time. '
                  '[WARNING]')

        if int_auto is not None:
            for i, item in enumerate(int_auto):
                dset.attrs[f'autocorrelation{i}'] = float(item)

        h5_file.close()

    @typechecked
    def get_probable_sample(self,
                            tag: str,
                            burnin: Optional[int] = None) -> Dict[str, float]:
        """
        Function for extracting the sample parameters with the highest posterior probability.

        Parameters
        ----------
        tag : str
            Database tag with the posterior results.
        burnin : int, None
            Number of burnin steps. No burnin is removed if set to ``None``.

        Returns
        -------
        dict
            Parameters and values for the sample with the maximum posterior probability.
        """

        if burnin is None:
            burnin = 0

        h5_file = h5py.File(self.database, 'r')
        dset = h5_file[f'results/fit/{tag}/samples']

        samples = np.asarray(dset)
        ln_prob = np.asarray(h5_file[f'results/fit/{tag}/ln_prob'])

        if 'n_param' in dset.attrs:
            n_param = dset.attrs['n_param']
        elif 'nparam' in dset.attrs:
            n_param = dset.attrs['nparam']

        if samples.ndim == 3:
            if burnin > samples.shape[1]:
                raise ValueError(f'The \'burnin\' value is larger than the number of steps '
                                 f'({samples.shape[1]}) that are made by the walkers.')

            samples = samples[:, burnin:, :]
            ln_prob = ln_prob[:, burnin:]

            samples = np.reshape(samples, (-1, n_param))
            ln_prob = np.reshape(ln_prob, -1)

        index_max = np.unravel_index(ln_prob.argmax(), ln_prob.shape)

        # max_prob = ln_prob[index_max]
        max_sample = samples[index_max]

        prob_sample = {}

        for i in range(n_param):
            par_key = dset.attrs[f'parameter{i}']
            par_value = max_sample[i]

            prob_sample[par_key] = par_value

        if dset.attrs.__contains__('distance'):
            prob_sample['distance'] = dset.attrs['distance']

        h5_file.close()

        return prob_sample

    @typechecked
    def get_median_sample(self,
                          tag: str,
                          burnin: Optional[int] = None) -> Dict[str, float]:
        """
        Function for extracting the median parameter values from the posterior samples.

        Parameters
        ----------
        tag : str
            Database tag with the posterior results.
        burnin : int, None
            Number of burnin steps. No burnin is removed if set to ``None``.

        Returns
        -------
        dict
            Parameters and values for the sample with the maximum posterior probability.
        """

        if burnin is None:
            burnin = 0

        with h5py.File(self.database, 'r') as h5_file:
            dset = h5_file[f'results/fit/{tag}/samples']

            if 'n_param' in dset.attrs:
                n_param = dset.attrs['n_param']
            elif 'nparam' in dset.attrs:
                n_param = dset.attrs['nparam']

            samples = np.asarray(dset)

            # samples = samples[samples[:, 2] > 100., ]

            if samples.ndim == 3:
                if burnin > samples.shape[1]:
                    raise ValueError(f'The \'burnin\' value is larger than the number of steps '
                                     f'({samples.shape[1]}) that are made by the walkers.')

                if burnin is not None:
                    samples = samples[:, burnin:, :]

                samples = np.reshape(samples, (-1, n_param))

            median_sample = {}

            for i in range(n_param):
                par_key = dset.attrs[f'parameter{i}']
                par_value = np.percentile(samples[:, i], 50.)
                median_sample[par_key] = par_value

            if dset.attrs.__contains__('distance'):
                median_sample['distance'] = dset.attrs['distance']

        return median_sample

    @typechecked
    def get_mcmc_spectra(self,
                         tag: str,
                         random: int,
                         burnin: Optional[int] = None,
                         wavel_range: Optional[Union[Tuple[float, float], str]] = None,
                         spec_res: Optional[float] = None) -> Union[List[box.ModelBox],
                                                                    List[box.SpectrumBox]]:
        """
        Parameters
        ----------
        tag : str
            Database tag with the posterior samples.
        random : int
            Number of random samples.
        burnin : int, None
            Number of burnin steps. No burnin is removed if set to ``None``.
        wavel_range : tuple(float, float), str, None
            Wavelength range (um) or filter name. Full spectrum is used if set to ``None``.
        spec_res : float
            Spectral resolution that is used for the smoothing with a Gaussian kernel. No smoothing
            is applied if set to ``None``.

        Returns
        -------
        list(species.core.box.ModelBox)
            Boxes with the randomly sampled spectra.
        """

        if burnin is None:
            burnin = 0

        h5_file = h5py.File(self.database, 'r')
        dset = h5_file[f'results/fit/{tag}/samples']

        spectrum_type = dset.attrs['type']
        spectrum_name = dset.attrs['spectrum']

        if 'n_param' in dset.attrs:
            n_param = dset.attrs['n_param']
        elif 'nparam' in dset.attrs:
            n_param = dset.attrs['nparam']

        if 'n_scaling' in dset.attrs:
            n_scaling = dset.attrs['n_scaling']
        elif 'nscaling' in dset.attrs:
            n_scaling = dset.attrs['nscaling']
        else:
            n_scaling = 0

        if 'n_error' in dset.attrs:
            n_error = dset.attrs['n_error']
        else:
            n_error = 0

        scaling = []
        for i in range(n_scaling):
            scaling.append(dset.attrs[f'scaling{i}'])

        error = []
        for i in range(n_error):
            error.append(dset.attrs[f'error{i}'])

        if spec_res is not None and spectrum_type == 'calibration':
            warnings.warn('Smoothing of the spectral resolution is not implemented for calibration '
                          'spectra.')

        if dset.attrs.__contains__('distance'):
            distance = dset.attrs['distance']
        else:
            distance = None

        samples = np.asarray(dset)

        # samples = samples[samples[:, 2] > 100., ]

        if samples.ndim == 2:
            ran_index = np.random.randint(samples.shape[0], size=random)
            samples = samples[ran_index, ]

        elif samples.ndim == 3:
            if burnin > samples.shape[1]:
                raise ValueError(f'The \'burnin\' value is larger than the number of steps '
                                 f'({samples.shape[1]}) that are made by the walkers.')

            samples = samples[:, burnin:, :]

            ran_walker = np.random.randint(samples.shape[0], size=random)
            ran_step = np.random.randint(samples.shape[1], size=random)
            samples = samples[ran_walker, ran_step, :]

        param = []
        for i in range(n_param):
            param.append(dset.attrs[f'parameter{i}'])

        if spectrum_type == 'model':
            if spectrum_name == 'planck':
                readmodel = read_planck.ReadPlanck(wavel_range)

            elif spectrum_name == 'powerlaw':
                pass

            else:
                readmodel = read_model.ReadModel(spectrum_name, wavel_range=wavel_range)

        elif spectrum_type == 'calibration':
            readcalib = read_calibration.ReadCalibration(spectrum_name, filter_name=None)

        boxes = []

        for i in tqdm.tqdm(range(samples.shape[0]), desc='Getting MCMC spectra'):
            model_param = {}
            for j in range(samples.shape[1]):
                if param[j] not in scaling and param[j] not in error:
                    model_param[param[j]] = samples[i, j]

            if distance:
                model_param['distance'] = distance

            if spectrum_type == 'model':
                if spectrum_name == 'planck':
                    specbox = readmodel.get_spectrum(model_param, spec_res)

                elif spectrum_name == 'powerlaw':
                    specbox = read_util.powerlaw_spectrum(wavel_range, model_param)

                else:
                    specbox = readmodel.get_model(model_param, spec_res=spec_res, smooth=True)

            elif spectrum_type == 'calibration':
                specbox = readcalib.get_spectrum(model_param)

            boxes.append(specbox)

        h5_file.close()

        return boxes

    @typechecked
    def get_mcmc_photometry(self,
                            tag: str,
                            filter_name: str,
                            burnin: Optional[int] = None) -> np.ndarray:
        """
        Parameters
        ----------
        tag : str
            Database tag with the posterior samples.
        filter_name : str
            Filter name for which the photometry will be computed.
        burnin : int, None
            Number of burnin steps. No burnin is removed if set to ``None``.

        Returns
        -------
        np.ndarray
            Synthetic magnitudes.
        """

        if burnin is None:
            burnin = 0

        h5_file = h5py.File(self.database, 'r')
        dset = h5_file[f'results/fit/{tag}/samples']

        if 'n_param' in dset.attrs:
            n_param = dset.attrs['n_param']
        elif 'nparam' in dset.attrs:
            n_param = dset.attrs['nparam']

        spectrum_type = dset.attrs['type']
        spectrum_name = dset.attrs['spectrum']

        if dset.attrs.__contains__('distance'):
            distance = dset.attrs['distance']
        else:
            distance = None

        samples = np.asarray(dset)

        if samples.ndim == 3:
            if burnin > samples.shape[1]:
                raise ValueError(f'The \'burnin\' value is larger than the number of steps '
                                 f'({samples.shape[1]}) that are made by the walkers.')

            samples = samples[:, burnin:, :]
            samples = samples.reshape((samples.shape[0]*samples.shape[1], n_param))

        param = []
        for i in range(n_param):
            param.append(dset.attrs[f'parameter{i}'])

        h5_file.close()

        if spectrum_type == 'model':
            if spectrum_name == 'powerlaw':
                synphot = photometry.SyntheticPhotometry(filter_name)
                synphot.zero_point()  # Set the wavel_range attribute

            else:
                readmodel = read_model.ReadModel(spectrum_name, filter_name=filter_name)

        elif spectrum_type == 'calibration':
            readcalib = read_calibration.ReadCalibration(spectrum_name, filter_name=filter_name)

        mcmc_phot = np.zeros((samples.shape[0]))

        for i in tqdm.tqdm(range(samples.shape[0]), desc='Getting MCMC photometry'):
            model_param = {}
            for j in range(n_param):
                model_param[param[j]] = samples[i, j]

            if distance is not None:
                model_param['distance'] = distance

            if spectrum_type == 'model':
                if spectrum_name == 'powerlaw':
                    pl_box = read_util.powerlaw_spectrum(synphot.wavel_range, model_param)

                    app_mag, _ = synphot.spectrum_to_magnitude(pl_box.wavelength, pl_box.flux)

                    mcmc_phot[i] = app_mag[0]

                else:
                    mcmc_phot[i], _ = readmodel.get_magnitude(model_param)

            elif spectrum_type == 'calibration':
                app_mag, _ = readcalib.get_magnitude(model_param=model_param, distance=None)
                mcmc_phot[i] = app_mag[0]

        return mcmc_phot

    @typechecked
    def get_object(self,
                   object_name: str,
                   inc_phot: Union[bool, List[str]] = True,
                   inc_spec: Union[bool, List[str]] = True) -> box.ObjectBox:
        """
        Function for extracting the photometric and/or spectroscopic data of an object from the
        database. The spectroscopic data contains optionally the covariance matrix and its inverse.

        Parameters
        ----------
        object_name : str
            Object name in the database.
        inc_phot : bool, list(str)
            Include photometric data. If a boolean, either all (``True``) or none (``False``) of
            the data are selected. If a list, a subset of filter names (as stored in the database)
            can be provided.
        inc_spec : bool, list(str)
            Include spectroscopic data. If a boolean, either all (``True``) or none (``False``) of
            the data are selected. If a list, a subset of spectrum names (as stored in the database
            with :func:`~species.data.database.Database.add_object`) can be provided.

        Returns
        -------
        species.core.box.ObjectBox
            Box with the object's data.
        """

        print(f'Getting object: {object_name}...', end='', flush=True)

        h5_file = h5py.File(self.database, 'r')
        dset = h5_file[f'objects/{object_name}']

        distance = np.asarray(dset['distance'])

        if inc_phot:

            magnitude = {}
            flux = {}

            for observatory in dset.keys():
                if observatory not in ['distance', 'spectrum']:
                    for filter_name in dset[observatory]:
                        name = f'{observatory}/{filter_name}'

                        if isinstance(inc_phot, bool) or name in inc_phot:
                            magnitude[name] = dset[name][0:2]
                            flux[name] = dset[name][2:4]

            phot_filters = list(magnitude.keys())

        else:

            magnitude = None
            flux = None
            phot_filters = None

        if inc_spec and f'objects/{object_name}/spectrum' in h5_file:
            spectrum = {}

            for item in h5_file[f'objects/{object_name}/spectrum']:
                data_group = f'objects/{object_name}/spectrum/{item}'

                if isinstance(inc_spec, bool) or item in inc_spec:

                    if f'{data_group}/covariance' not in h5_file:
                        spectrum[item] = (np.asarray(h5_file[f'{data_group}/spectrum']),
                                          None,
                                          None,
                                          h5_file[f'{data_group}'].attrs['specres'])

                    else:
                        spectrum[item] = (np.asarray(h5_file[f'{data_group}/spectrum']),
                                          np.asarray(h5_file[f'{data_group}/covariance']),
                                          np.asarray(h5_file[f'{data_group}/inv_covariance']),
                                          h5_file[f'{data_group}'].attrs['specres'])

        else:
            spectrum = None

        h5_file.close()

        print(' [DONE]')

        return box.create_box('object',
                              name=object_name,
                              filters=phot_filters,
                              magnitude=magnitude,
                              flux=flux,
                              distance=distance,
                              spectrum=spectrum)

    @typechecked
    def get_samples(self,
                    tag: str,
                    burnin: Optional[int] = None,
                    random: Optional[int] = None) -> box.SamplesBox:
        """
        Parameters
        ----------
        tag: str
            Database tag with the samples.
        burnin : int, None
            Number of burnin samples to exclude. All samples are selected if set to ``None``.
            The parameter is only required for samples obtained with ``emcee`` and is therefore
            not used for samples obtained with ``MultiNest``.
        random : int, None
            Number of random samples to select. All samples (with the burnin excluded) are
            selected if set to ``None``.

        Returns
        -------
        species.core.box.SamplesBox
            Box with the posterior samples.
        """

        if burnin is None:
            burnin = 0

        h5_file = h5py.File(self.database, 'r')
        dset = h5_file[f'results/fit/{tag}/samples']
        ln_prob = np.asarray(h5_file[f'results/fit/{tag}/ln_prob'])

        spectrum = dset.attrs['spectrum']

        if 'n_param' in dset.attrs:
            n_param = dset.attrs['n_param']
        elif 'nparam' in dset.attrs:
            n_param = dset.attrs['nparam']

        samples = np.asarray(dset)

        if samples.ndim == 3:
            if burnin > samples.shape[1]:
                raise ValueError(f'The \'burnin\' value is larger than the number of steps '
                                 f'({samples.shape[1]}) that are made by the walkers.')

            samples = samples[:, burnin:, :]

            if random:
                ran_walker = np.random.randint(samples.shape[0], size=random)
                ran_step = np.random.randint(samples.shape[1], size=random)
                samples = samples[ran_walker, ran_step, :]

        param = []
        for i in range(n_param):
            param.append(dset.attrs[f'parameter{i}'])

        h5_file.close()

        if samples.ndim == 3:
            prob_sample = self.get_probable_sample(tag, burnin)
        else:
            prob_sample = None

        median_sample = self.get_median_sample(tag, burnin)

        return box.create_box('samples',
                              spectrum=spectrum,
                              parameters=param,
                              samples=samples,
                              ln_prob=ln_prob,
                              prob_sample=prob_sample,
                              median_sample=median_sample)



import os
import shutil

import pytest
import numpy as np

import species
from species.util import test_util


class TestCalibration:

    def setup_class(self):
        self.limit = 1e-8
        self.test_path = os.path.dirname(__file__) + '/'
        self.model_param = {'scaling': 1.}

    def teardown_class(self):
        os.remove('species_database.hdf5')
        os.remove('species_config.ini')
        shutil.rmtree('data/')

    def test_species_init(self):
        test_util.create_config('./')
        species.SpeciesInit()

    def test_read_calibration(self):
        database = species.Database()
        database.add_spectrum('vega')

        read_calib = species.ReadCalibration('vega', filter_name='Paranal/NACO.H')
        assert read_calib.wavel_range == pytest.approx((1.44, 1.88), rel=1e-7, abs=0.)

    def test_resample_spectrum(self):
        read_calib = species.ReadCalibration('vega')
        spec_box = read_calib.resample_spectrum(np.linspace(1., 2., 10), apply_mask=True)

        assert np.sum(spec_box.wavelength) == 15.
        assert np.sum(spec_box.flux) == pytest.approx(2.288734760321133e-08, rel=self.limit, abs=0.)

    def test_get_spectrum(self):
        read_calib = species.ReadCalibration('vega', filter_name='Paranal/NACO.Lp')
        spec_box = read_calib.get_spectrum(self.model_param, apply_mask=True, spec_res=100.)

        assert np.sum(spec_box.wavelength) == pytest.approx(183.34793783628004, rel=self.limit, abs=0.)
        assert np.sum(spec_box.flux) == pytest.approx(2.363239096767195e-09, rel=self.limit, abs=0.)

    def test_get_flux(self):
        read_calib = species.ReadCalibration('vega', filter_name='Paranal/NACO.H')
        flux = read_calib.get_flux(model_param=self.model_param)

        assert flux[0]  == pytest.approx(1.132902304231636e-09, rel=self.limit, abs=0.)

    def test_get_magnitude(self):
        read_calib = species.ReadCalibration('vega', filter_name='Paranal/NACO.H')
        app_mag, abs_mag = read_calib.get_magnitude(model_param=self.model_param)

        assert app_mag[0] == 0.03
        assert abs_mag[0] == None



"""
Utility functions for plotting data.
"""

from typing import Optional, Tuple, List

import numpy as np

from typeguard import typechecked


def sptype_substellar(sptype,
                      shape):
    """
    Parameters
    ----------
    sptype :
    shape :

    Returns
    -------
    numpy.ndarray
    """

    spt_disc = np.zeros(shape)

    for i, item in enumerate(sptype):
        if item[0:2] in ['M0', 'M1', 'M2', 'M3', 'M4']:
            spt_disc[i] = 0.5

        elif item[0:2] in ['M5', 'M6', 'M7', 'M8', 'M9']:
            spt_disc[i] = 1.5

        elif item[0:2] in ['L0', 'L1', 'L2', 'L3', 'L4']:
            spt_disc[i] = 2.5

        elif item[0:2] in ['L5', 'L6', 'L7', 'L8', 'L9']:
            spt_disc[i] = 3.5

        elif item[0:2] in ['T0', 'T1', 'T2', 'T3', 'T4']:
            spt_disc[i] = 4.5

        elif item[0:2] in ['T5', 'T6', 'T7', 'T8', 'T9']:
            spt_disc[i] = 5.5

        elif 'Y' in item:
            spt_disc[i] = 6.5

        else:
            spt_disc[i] = np.nan
            continue

    return spt_disc


def sptype_stellar(sptype,
                   shape):
    """
    Parameters
    ----------
    sptype :
    shape :

    Returns
    -------
    numpy.ndarray
    """

    spt_disc = np.zeros(shape)

    for i, item in enumerate(sptype):
        if item[0] == 'O':
            spt_disc[i] = 0.5

        elif item[0] == 'B':
            spt_disc[i] = 1.5

        elif item[0] == 'A':
            spt_disc[i] = 2.5

        elif item[0] == 'F':
            spt_disc[i] = 3.5

        elif item[0] == 'G':
            spt_disc[i] = 4.5

        elif item[0] == 'K':
            spt_disc[i] = 5.5

        elif item[0] == 'M':
            spt_disc[i] = 6.5

        elif item[0] == 'L':
            spt_disc[i] = 7.5

        elif item[0] == 'T':
            spt_disc[i] = 8.5

        elif item[0] == 'Y':
            spt_disc[i] = 9.5

        else:
            spt_disc[i] = np.nan
            continue

    return spt_disc


@typechecked
def update_labels(param: List[str]) -> List[str]:
    """
    Function for updating the fitted parameters to labels used in plots.

    Parameters
    ----------
    param : list
        List with parameter names.

    Returns
    -------
    list
        List with parameter labels for plots.
    """

    if 'teff' in param:
        index = param.index('teff')
        param[index] = r'$\mathregular{T}_\mathregular{eff}$ (K)'

    if 'logg' in param:
        index = param.index('logg')
        param[index] = r'$\mathregular{log}\,\mathregular{g}$'

    if 'feh' in param:
        index = param.index('feh')
        param[index] = r'[Fe/H]'

    if 'fsed' in param:
        index = param.index('fsed')
        param[index] = r'$\mathregular{f}_\mathregular{sed}$'

    if 'co' in param:
        index = param.index('co')
        param[index] = r'C/O'

    if 'radius' in param:
        index = param.index('radius')
        param[index] = r'$\mathregular{R}$ ($\mathregular{R_J}$)'

    if 'mass' in param:
        index = param.index('mass')
        param[index] = r'$\mathregular{M}$ ($\mathregular{M_J}$)'

    if 'luminosity' in param:
        index = param.index('luminosity')
        param[index] = r'$\mathregular{log}\,\mathregular{L}/\mathregular{L}_\mathregular{\odot}$'

    if 'luminosity_ratio' in param:
        index = param.index('luminosity_ratio')
        param[index] = r'$\mathregular{log}\,\mathregular{L_1}/\mathregular{L_2}$'

    if 'lognorm_radius' in param:
        index = param.index('lognorm_radius')
        param[index] = r'$\mathregular{log}\,\mathregular{r_g}$'

    if 'lognorm_sigma' in param:
        index = param.index('lognorm_sigma')
        param[index] = r'$\mathregular{\sigma_g}$'

    if 'lognorm_ext' in param:
        index = param.index('lognorm_ext')
        param[index] = r'$\mathregular{A_V}$'

    if 'powerlaw_min' in param:
        index = param.index('powerlaw_min')
        param[index] = r'$\mathregular{{log}}\,\mathregular{a}_\mathregular{min}/\mathregular{µm}$'

    if 'powerlaw_max' in param:
        index = param.index('powerlaw_max')
        param[index] = r'$\mathregular{{log}}\,\mathregular{a}_\mathregular{max}/\mathregular{µm}$'

    if 'powerlaw_exp' in param:
        index = param.index('powerlaw_exp')
        param[index] = r'$\beta$'

    if 'powerlaw_ext' in param:
        index = param.index('powerlaw_ext')
        param[index] = r'$\mathregular{A_V}$'

    if 'ism_ext' in param:
        index = param.index('ism_ext')
        param[index] = r'$\mathregular{A_V}$'

    if 'ism_red' in param:
        index = param.index('ism_red')
        param[index] = r'$\mathregular{R_V}$'

    if 'tint' in param:
        index = param.index('tint')
        param[index] = r'$\mathregular{T_{int}}$ (K)'

    for i in range(15):
        if f't{i}' in param:
            index = param.index(f't{i}')
            param[index] = rf'$\mathregular{{T}}_\mathregular{{{i}}}$ (K)'

    if 'alpha' in param:
        index = param.index('alpha')
        param[index] = r'$\alpha$'

    if 'log_delta' in param:
        index = param.index('log_delta')
        param[index] = r'$\mathregular{log}\,\mathregular{\delta}$'

    if 'log_p_quench' in param:
        index = param.index('log_p_quench')
        param[index] = r'$\mathregular{log}\,\mathregular{P}_\mathregular{quench}$'

    for i, item in enumerate(param):
        if item[0:8] == 'scaling_':
            param[i] = rf'$\mathregular{{a}}_\mathregular{{{item[8:]}}}$'

        elif item[0:6] == 'error_':
            param[i] = rf'$\mathregular{{b}}_\mathregular{{{item[6:]}}}$'

        elif item[0:11] == 'wavelength_':
            param[i] = rf'$\mathregular{{c}}_\mathregular{{{item[11:]}}}$ (nm)'

        elif item[0:9] == 'corr_len_':
            param[i] = rf'$\mathregular{{log}}\,\ell_\mathregular{{{item[9:]}}}$'

        elif item[0:9] == 'corr_amp_':
            param[i] = rf'$\mathregular{{f}}_\mathregular{{{item[9:]}}}$'

    for i in range(100):
        if f'teff_{i}' in param:
            index = param.index(f'teff_{i}')
            param[index] = rf'$\mathregular{{T}}_\mathregular{{{i+1}}}$ (K)'

        else:
            break

    for i in range(100):
        if f'radius_{i}' in param:
            index = param.index(f'radius_{i}')
            param[index] = rf'$\mathregular{{R}}_\mathregular{{{i+1}}}$ ' + r'($\mathregular{R_J}$)'

        else:
            break

    for i in range(100):
        if f'luminosity_{i}' in param:
            index = param.index(f'luminosity_{i}')
            param[index] = rf'$\mathregular{{log}}\,\mathregular{{L}}_\mathregular{{{i+1}}}' \
                           rf'/\mathregular{{L}}_\mathregular{{\odot}}$'

        else:
            break

    if 'disk_teff' in param:
        index = param.index('disk_teff')
        param[index] = r'$\mathregular{T}_\mathregular{disk}$ (K)'

    if 'disk_radius' in param:
        index = param.index('disk_radius')
        param[index] = r'$\mathregular{R}_\mathregular{disk}$ ($\mathregular{R_J}$)'

    if 'log_powerlaw_a' in param:
        index = param.index('log_powerlaw_a')
        param[index] = r'$a_\mathregular{powerlaw}$'

    if 'log_powerlaw_b' in param:
        index = param.index('log_powerlaw_b')
        param[index] = r'$b_\mathregular{powerlaw}$'

    if 'log_powerlaw_c' in param:
        index = param.index('log_powerlaw_c')
        param[index] = r'$c_\mathregular{powerlaw}$'

    return param


@typechecked
def model_name(key: str) -> str:
    """
    Function for updating a model name for use in plots.

    Parameters
    ----------
    key : str
        Model name as used by species.

    Returns
    -------
    str
        Updated model name for plots.
    """

    if key == 'drift-phoenix':
        name = 'DRIFT-PHOENIX'

    elif key == 'ames-cond':
        name = 'AMES-Cond'

    elif key == 'ames-dusty':
        name = 'AMES-Dusty'

    elif key == 'atmo':
        name = 'ATMO'

    elif key == 'bt-cond':
        name = 'BT-Cond'

    elif key == 'bt-cond-feh':
        name = 'BT-Cond'

    elif key == 'bt-settl':
        name = 'BT-Settl'

    elif key == 'bt-settl-cifist':
        name = 'BT-Settl'

    elif key == 'bt-nextgen':
        name = 'BT-NextGen'

    elif key == 'petitcode-cool-clear':
        name = 'petitCODE'

    elif key == 'petitcode-cool-cloudy':
        name = 'petitCODE'

    elif key == 'petitcode-hot-clear':
        name = 'petitCODE'

    elif key == 'petitcode-hot-cloudy':
        name = 'petitCODE'

    elif key == 'exo-rem':
        name = 'Exo-REM'

    elif key == 'planck':
        name = 'Blackbody radiation'

    elif key == 'zhu2015':
        name = 'Zhu (2015)'

    return name


@typechecked
def quantity_unit(param: List[str],
                  object_type: str) -> Tuple[List[str], List[Optional[str]], List[str]]:
    """
    Function for creating lists with quantities, units, and labels for fitted parameter.

    Parameters
    ----------
    param : list
        List with parameter names.
    object_type : str
        Object type (``'planet'`` or ``'star'``).

    Returns
    -------
    list
        List with the quantities.
    list
        List with the units.
    list
        List with the parameter labels for plots.
    """

    quantity = []
    unit = []
    label = []

    if 'teff' in param:
        quantity.append('teff')
        unit.append('K')
        label.append(r'$\mathregular{T}_\mathregular{eff}$')

    if 'logg' in param:
        quantity.append('logg')
        unit.append(None)
        label.append(r'$\mathregular{log}\,\mathregular{g}$')

    if 'feh' in param:
        quantity.append('feh')
        unit.append(None)
        label.append('[Fe/H]')

    if 'fsed' in param:
        quantity.append('fsed')
        unit.append(None)
        label.append(r'$\mathregular{f}_\mathregular{sed}$')

    if 'co' in param:
        quantity.append('co')
        unit.append(None)
        label.append('C/O')

    if 'radius' in param:
        quantity.append('radius')

        if object_type == 'planet':
            unit.append(r'$\mathregular{R}_\mathregular{J}$')

        elif object_type == 'star':
            unit.append(r'$\mathregular{R}_\mathregular{\odot}$')

        label.append(r'$\mathregular{R}$')

    for i in range(100):
        if f'teff_{i}' in param:
            quantity.append(f'teff_{i}')
            unit.append('K')
            label.append(rf'$\mathregular{{T}}_\mathregular{{{i+1}}}$')

        else:
            break

    for i in range(100):
        if f'radius_{i}' in param:
            quantity.append(f'radius_{i}')
            unit.append(r'$\mathregular{R}_\mathregular{J}$')
            label.append(rf'$\mathregular{{R}}_\mathregular{{{i+1}}}$')

        else:
            break

    if 'distance' in param:
        quantity.append('distance')
        unit.append('pc')
        label.append(r'$\mathregular{d}$')

    if 'mass' in param:
        quantity.append('mass')

        if object_type == 'planet':
            unit.append(r'$\mathregular{M}_\mathregular{J}$')

        elif object_type == 'star':
            unit.append(r'$\mathregular{M}_\mathregular{\odot}$')

        label.append('M')

    if 'luminosity' in param:
        quantity.append('luminosity')
        unit.append(None)
        label.append(r'$\mathregular{log}\,\mathregular{L}/\mathregular{L}_\mathregular{\odot}$')

    if 'ism_ext' in param:
        quantity.append('ism_ext')
        unit.append(None)
        label.append(r'$\mathregular{A}_\mathregular{V}$')

    if 'lognorm_ext' in param:
        quantity.append('lognorm_ext')
        unit.append(None)
        label.append(r'$\mathregular{A}_\mathregular{V}$')

    if 'powerlaw_ext' in param:
        quantity.append('powerlaw_ext')
        unit.append(None)
        label.append(r'$\mathregular{A}_\mathregular{V}$')

    return quantity, unit, label


def field_bounds_ticks(field_range):
    """
    Parameters
    ----------
    field_range : tuple(str, str), None
        Range of the discrete colorbar for the field dwarfs. The tuple should contain the lower
        and upper value ('early M', 'late M', 'early L', 'late L', 'early T', 'late T', 'early Y).
        The full range is used if set to None.

    Returns
    -------
    np.ndarray
    np.ndarray
    list(str, )
    """

    spectral_ranges = ['M0-M4', 'M5-M9', 'L0-L4', 'L5-L9', 'T0-T4', 'T5-T9', 'Y1-Y2']

    if field_range is None:
        index_start = 0
        index_end = 7

    else:
        if field_range[0] == 'early M':
            index_start = 0
        elif field_range[0] == 'late M':
            index_start = 1
        elif field_range[0] == 'early L':
            index_start = 2
        elif field_range[0] == 'late L':
            index_start = 3
        elif field_range[0] == 'early T':
            index_start = 4
        elif field_range[0] == 'late T':
            index_start = 5
        elif field_range[0] == 'early Y':
            index_start = 6

        if field_range[1] == 'early M':
            index_end = 1
        elif field_range[1] == 'late M':
            index_end = 2
        elif field_range[1] == 'early L':
            index_end = 3
        elif field_range[1] == 'late L':
            index_end = 4
        elif field_range[1] == 'early T':
            index_end = 5
        elif field_range[1] == 'late T':
            index_end = 6
        elif field_range[1] == 'early Y':
            index_end = 7

    index_range = index_end - index_start + 1

    bounds = np.linspace(index_start, index_end, index_range)
    ticks = np.linspace(index_start+0.5, index_end-0.5, index_range-1)
    labels = spectral_ranges[index_start:index_end]

    return bounds, ticks, labels


@typechecked
def remove_color_duplicates(object_names: List[str],
                            empirical_names: np.ndarray) -> List[int]:
    """"
    Function for deselecting young/low-gravity objects that will already be plotted individually
    as directly imaged objects.

    Parameters
    ----------
    object_names : list(str)
        List with names of directly imaged planets and brown dwarfs.
    empirical_names : np.ndarray
        Array with names of young/low-gravity objects.

    Returns
    -------
    list
        List with selected indices of the young/low-gravity objects.
    """

    indices = []

    for i, item in enumerate(empirical_names):
        if item == 'beta_Pic_b' and 'beta Pic b' in object_names:
            continue

        if item == 'HR8799b' and 'HR 8799 b' in object_names:
            continue

        if item == 'HR8799c' and 'HR 8799 c' in object_names:
            continue

        if item == 'HR8799d' and 'HR 8799 d' in object_names:
            continue

        if item == 'HR8799e' and 'HR 8799 e' in object_names:
            continue

        if item == 'kappa_And_B' and 'kappa And b' in object_names:
            continue

        if item == 'HD1160B' and 'HD 1160 B' in object_names:
            continue

        indices.append(i)

    return indices
